
import tkinter as tk
import tkinter.ttk
import tkinter.simpledialog
import tkinter.messagebox
import logging

logger = logging.getLogger(__name__)


class RunnerSavePrompt(tkinter.simpledialog.Dialog):
    """
    Dialog box for prompting the user to save before running.
    """

    def __init__(self, parent_window, title="Enrich2"):
        self.pw = parent_window

        self.dialog_text = tk.StringVar()
        self.dialog_text.set("Would you like to save your config changes?")

        tkinter.simpledialog.Dialog.__init__(self, parent_window, title)

    def body(self, master):
        frame = tkinter.ttk.Frame(master, padding=(12, 6, 12, 6))
        frame.pack()

        dialog_text_label = tkinter.ttk.Label(frame, textvariable=self.dialog_text)
        dialog_text_label.grid(column=0, row=0, sticky="nsew")

    def apply(self):
        self.pw.menu_save()


class RunnerWindow(tkinter.simpledialog.Dialog):
    """
    Dialog box for blocking input while running the analysis.
    """

    def __init__(self, parent_window, title="Enrich2"):
        self.pw = parent_window
        self.run_button = None

        self.dialog_text = tk.StringVar()
        self.dialog_text.set("Ready to start analysis...")

        tkinter.simpledialog.Dialog.__init__(self, parent_window, title)

    def body(self, master):
        frame = tkinter.ttk.Frame(master, padding=(12, 6, 12, 6))
        frame.pack()

        dialog_text_label = tkinter.ttk.Label(frame, textvariable=self.dialog_text)
        dialog_text_label.grid(column=0, row=0, sticky="nsew")

        self.run_button = tk.Button(
            frame, text="Begin", width=10, command=self.runner, default="active"
        )
        self.run_button.grid(column=0, row=1, sticky="nsew")

    def buttonbox(self):
        """
        Display no buttons.
        """
        pass

    def runner(self):
        # gray out the "Run" button
        self.run_button.config(state="disabled")
        self.update_idletasks()

        # set the analysis options
        self.pw.root_element.force_recalculate = self.pw.force_recalculate.get()
        self.pw.root_element.component_outliers = self.pw.component_outliers.get()
        self.pw.root_element.scoring_method = self.pw.scoring_method.get()
        self.pw.root_element.logr_method = self.pw.logr_method.get()
        self.pw.root_element.plots_requested = self.pw.plots_requested.get()
        self.pw.root_element.tsv_requested = self.pw.tsv_requested.get()

        # run the analysis, catching any errors to display in a dialog box
        try:
            # ensure that all objects are valid
            self.pw.root_element.validate()

            # open HDF5 files for the root and all child objects
            self.pw.root_element.store_open(children=True)

            # perform the analysis
            self.pw.root_element.calculate()

        except Exception as e:
            # display error
            logger.error(e)
            tkinter.messagebox.showerror(
                "Enrich2 Error", "Enrich2 encountered an error:\n{}".format(e)
            )

        else:
            # no exception occurred during calculation and setup
            # generate desired output
            if self.pw.plots_requested.get():
                try:
                    self.pw.root_element.make_plots()
                except Exception as e:
                    tkinter.messagebox.showwarning(
                        None,
                        "Calculations completed, but plotting failed:\n{}".format(e),
                    )
            if self.pw.tsv_requested.get():
                try:
                    self.pw.root_element.write_tsv()
                except Exception as e:
                    tkinter.messagebox.showwarning(
                        None,
                        "Calculations completed, but tsv output failed:\n{}".format(e),
                    )

            # show the dialog box
            tkinter.messagebox.showinfo("", "Analysis completed.")

        finally:
            # close the HDF5 files
            self.pw.root_element.store_close(children=True)

            # close this window
            self.destroy()



import re
import gzip
import bz2
import os.path

re_barcode = re.compile("^[ACGT]+$")
re_variant_dna = re.compile("^[ACGTN]+$")
re_identifier = re.compile("^.+$")


class BarcodeMap(dict):
    """
    Dictionary-derived class for storing the relationship between barcodes
    (keys) and variants (values). Requires the path to a *mapfile*, containing
    lines in the format ``'barcode<whitespace>variant'`` for each barcode
    expected in the library. This file can be plain text or compressed
    (``.bz2`` or ``.gz``).

    Barcodes must only contain the characters ``ACGT`` and variants must only
    contain the characters ``ACGTN`` (lowercase characters are converted to
    uppercase).

    Blank lines and lines that begin with ``#`` (comments) are ignored.

    *is_variant* is a boolean that is ``True`` if the barcodes are assigned to
    variant DNA sequences, or ``False`` if the barcodes are assigned to
    arbitrary identifiers. If this is ``True``, additional error checking
    is performed on the variant DNA sequences.

    """

    def __init__(self, mapfile, is_variant=False):
        super(BarcodeMap, self).__init__()
        self.name = "barcodemap_{}".format(os.path.basename(mapfile))
        self.filename = mapfile
        self.is_variant = is_variant

        # open the file
        try:
            ext = os.path.splitext(mapfile)[-1].lower()
            if ext in (".bz2",):
                handle = bz2.BZ2File(mapfile, "r")
            elif ext in (".gz",):
                handle = gzip.GzipFile(mapfile, "r")
            else:
                handle = open(mapfile, "r")
        except IOError:
            raise IOError(
                "Could not open barcode map file '{}' [{}]".format(mapfile, self.name)
            )

        # handle each line
        for line in handle:
            line = line.decode("utf-8")
            # skip comments and whitespace-only lines
            if len(line.strip()) == 0 or line[0] == "#":
                continue

            try:
                barcode, value = line.strip().split()
            except ValueError:
                raise ValueError(
                    "Unexpected barcode map line format " "[{}]".format(self.name)
                )

            barcode = barcode.upper()
            if not re_barcode.match(barcode):
                raise ValueError(
                    "Barcode DNA sequence contains unexpected "
                    "characters [{}]".format(self.name)
                )
            if self.is_variant:
                value = value.upper()
                if not re_variant_dna.match(value):
                    raise ValueError(
                        "Variant DNA sequence contains unexpected"
                        " characters [{}]".format(self.name)
                    )
            else:
                if not re_identifier.match(value):
                    raise ValueError(
                        "Identifier contains unexpected "
                        "characters [{}]".format(self.name)
                    )

            if barcode in self:
                if self[barcode] != value:
                    raise ValueError(
                        "Barcode '{}' assigned to multiple "
                        "unique values: {}".format(barcode, self.name)
                    )
            else:
                self[barcode] = value

        handle.close()



import tkinter as tk
import tkinter.ttk
import tkinter.simpledialog


class SeqLibApplyDialog(tkinter.simpledialog.Dialog):
    """
    Confirmation dialog box for applying FASTQ filtering options to selected SeqLibs from the Treeview.
    """

    def __init__(
        self, parent_window, tree, source_id, title="Confirm Filtering Changes"
    ):
        self.tree = tree
        self.source_id = source_id
        self.target_ids = [
            x
            for x in self.tree.treeview.selection()
            if x != source_id
            and type(self.tree.get_element(self.source_id))
            == type(self.tree.get_element(x))
        ]
        tkinter.simpledialog.Dialog.__init__(self, parent_window, title)

    def body(self, master):
        """
        Generates the required text listing all SeqLibs that will have their FASTQ options updated.

        Displays the "OK" and "Cancel" buttons.
        """
        if len(self.target_ids) == 0:
            message_string = "No elegible SeqLibs selected."
        elif len(self.target_ids) == 1:
            message_string = 'Apply FASTQ filtering options from "{}" to "{}"?'.format(
                self.tree.get_element(self.source_id).name,
                self.tree.get_element(self.target_ids[0]).name,
            )
        else:
            bullet = "    " + "\u25C6"
            message_string = 'Apply FASTQ filtering options from "{}"" to the following?\n'.format(
                self.tree.get_element(self.source_id).name
            )
            for x in self.target_ids:
                message_string += "{bullet} {name}\n".format(
                    bullet=bullet, name=self.tree.get_element(x).name
                )
        message = tkinter.ttk.Label(master, text=message_string, justify="left")
        message.grid(row=0, sticky="w")

    def buttonbox(self):
        """
        Display only one button if there's no selection. Otherwise, use the default method to display two buttons.
        """
        if len(self.target_ids) == 0:
            box = tk.Frame(self)

            w = tk.Button(
                box, text="OK", width=10, command=self.cancel, default="active"
            )
            w.pack(side="left", padx=5, pady=5)

            self.bind("<Return>", self.cancel)

            box.pack()
        else:
            tkinter.simpledialog.Dialog.buttonbox(self)

    def apply(self):
        """
        Called when the user chooses "OK". Performs the FASTQ filtering update.
        """
        filter_cfg = self.tree.get_element(self.source_id).serialize_filters()
        for x in self.target_ids:
            self.tree.get_element(x).filters = filter_cfg
        self.tree.refresh_treeview()



#!/usr/bin/env python
#

from argparse import ArgumentParser, RawDescriptionHelpFormatter
import logging
import json
import sys
import platform
import os.path

if platform.system() == "Darwin":
    # Explicitly set the backend to avoid the NSInvalidArgumentException when
    # running in GUI mode. Advanced users who want to use another matplotlib
    # backend when running in MacOS on the command line can modify this section
    # accordingly.
    import matplotlib

    matplotlib.use("TkAgg")
elif os.path.exists("/.dockerenv"):
    # Explicitly set the backend for running inside Docker. This may fail for
    # older versions of docker or alternative containerization tools such as
    # Singularity.
    import matplotlib

    matplotlib.use("Agg")
import enrich2.config_check as config_check
from enrich2.experiment import Experiment
from enrich2.selection import Selection
from enrich2.barcode import BarcodeSeqLib
from enrich2.barcodeid import BcidSeqLib
from enrich2.barcodevariant import BcvSeqLib
from enrich2.basic import BasicSeqLib
from enrich2.overlap import OverlapSeqLib
from enrich2.idonly import IdOnlySeqLib
from enrich2.storemanager import SCORING_METHODS, LOGR_METHODS
from enrich2.gui.configurator import Configurator
from enrich2.sfmap import parse_aa_list
from enrich2 import __version__


#: Name of the driver script. Used for logging output.
DRIVER_NAME = os.path.basename(sys.argv[0])


#: Format string for log entries (console or file).
LOG_FORMAT = "%(asctime)-15s [%(name)s] %(message)s"

#: Default log level
LOG_LEVEL = logging.INFO


def main_gui():
    """
    Entry point for GUI.

    """
    logging.basicConfig(level=LOG_LEVEL, format=LOG_FORMAT)
    app = Configurator(__version__)
    app.mainloop()


def main_cmd():
    """
    Entry point for command line.

    """
    # build description string based on available methods
    desc_string = (
        "Command-line driver for Enrich2 v{}".format(__version__)
        + "\n\nscoring methods:\n"
        + "\n".join(["  {:22}{}".format(k, v) for k, v in list(SCORING_METHODS.items())])
        + "\n\nlog ratio methods:\n"
        + "\n".join(["  {:22}{}".format(k, v) for k, v in list(LOGR_METHODS.items())])
    )

    # create parser and add description
    parser = ArgumentParser(
        prog="Enrich2",
        description=desc_string,
        formatter_class=RawDescriptionHelpFormatter,
    )

    # add command line arguments
    parser.add_argument("config", help="JSON configuration file")
    parser.add_argument(
        "scoring_method", help="scoring method", choices=list(SCORING_METHODS.keys())
    )
    parser.add_argument(
        "logr_method", help="log ratio method", choices=list(LOGR_METHODS.keys())
    )

    # add support for semantic version checking
    parser.add_argument(
        "--version", action="version", version="{}".format(__version__)
    )

    # add analysis options
    parser.add_argument(
        "--log", metavar="FILE", dest="log_file", help="path to log file"
    )
    parser.add_argument(
        "--no-plots",
        dest="plots_requested",
        action="store_false",
        default=True,
        help="don't make plots",
    )
    parser.add_argument(
        "--no-tsv",
        dest="tsv_requested",
        action="store_false",
        default=True,
        help="don't generate tsv files",
    )
    parser.add_argument(
        "--recalculate",
        dest="force_recalculate",
        action="store_true",
        default=False,
        help="force recalculation",
    )
    parser.add_argument(
        "--component-outliers",
        dest="component_outliers",
        action="store_true",
        default=False,
        help="calculate component outlier stats",
    )
    parser.add_argument(
        "--output-dir",
        metavar="DIR",
        dest="output_dir_override",
        help="override the config file's output directory",
    )
    parser.add_argument(
        "--sfmap-aa-file",
        metavar="FILE",
        dest="sfmap_aa_file",
        help="amino acid groups for sequence-function maps",
    )

    args = parser.parse_args()

    # start the logs
    if args.log_file is not None:
        # Create directory if it doesn't exist
        log_dir = os.path.dirname(args.log_file)
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        logging.basicConfig(filename=args.log_file, encoding='utf-8', level=LOG_LEVEL, format=LOG_FORMAT)
    else:
        logging.basicConfig(level=LOG_LEVEL, format=LOG_FORMAT)
    logger = logging.getLogger(__name__)

    # read the JSON file
    try:
        cfg = json.load(open(args.config, "r"))
    except IOError:
        raise IOError("Failed to open '{}' [{}]".format(args.config, DRIVER_NAME))
    except ValueError:
        raise ValueError("Improperly formatted .json file [{}]".format(DRIVER_NAME))

    # identify config file type and create the object
    if config_check.is_experiment(cfg):
        logger.info("Detected an Experiment config file")
        obj = Experiment()
    elif config_check.is_selection(cfg):
        logger.info("Detected a Selection config file")
        obj = Selection()
    elif config_check.is_seqlib(cfg):
        seqlib_type = config_check.seqlib_type(cfg)
        logger.info("Detected a %s config file", seqlib_type)
        if seqlib_type == "BarcodeSeqLib":
            obj = BarcodeSeqLib()
        elif seqlib_type == "BcidSeqLib":
            obj = BcidSeqLib()
        elif seqlib_type == "BcvSeqLib":
            obj = BcvSeqLib()
        elif seqlib_type == "BasicSeqLib":
            obj = BasicSeqLib()
        elif seqlib_type == "OverlapSeqLib":
            obj = OverlapSeqLib()
        elif seqlib_type == "IdOnlySeqLib":
            obj = IdOnlySeqLib()
        else:
            raise ValueError(
                "Unrecognized SeqLib type '{}' [{}]".format(seqlib_type, DRIVER_NAME)
            )
    else:
        raise ValueError("Unrecognized .json config [{}]".format(DRIVER_NAME))

    # set analysis options
    obj.force_recalculate = args.force_recalculate
    obj.component_outliers = args.component_outliers
    obj.scoring_method = args.scoring_method
    obj.logr_method = args.logr_method
    obj.plots_requested = args.plots_requested
    obj.tsv_requested = args.tsv_requested

    if args.output_dir_override is not None:
        obj.output_dir_override = True
        obj.output_dir = args.output_dir_override
    else:
        obj.output_dir_override = False

    if args.sfmap_aa_file is not None:
        obj.plot_options = dict()
        obj.plot_options["aa_list"], obj.plot_options[
            "aa_label_groups"
        ] = parse_aa_list(args.sfmap_aa_file)

    # configure the object
    obj.configure(cfg)

    # make sure objects are valid
    try:
        obj.validate()
    except ValueError:
        logger.exception("Invalid configuration")
    else:
        # open HDF5 files for the object and all child objects
        obj.store_open(children=True)

        # perform the analysis
        obj.calculate()

        # generate desired output
        obj.make_plots()
        try:
            obj.make_plots()
        except Exception:
            logger.exception("Calculations completed, but plotting failed.")
        try:
            obj.write_tsv()
        except Exception:
            logger.exception("Calculations completed, but TSV output failed.")

        # clean up
        obj.store_close(children=True)


if __name__ == "__main__":
    gui_mode = False

    try:
        if sys.argv[1] == "gui":
            gui_mode = True
    except IndexError:
        pass

    if gui_mode:
        main_gui()
    else:
        main_cmd()




import tkinter as tk
import tkinter.ttk
import tkinter.simpledialog
from collections import OrderedDict
from ..barcode import BarcodeSeqLib
from ..barcodevariant import BcvSeqLib
from ..barcodeid import BcidSeqLib
from ..basic import BasicSeqLib
from ..idonly import IdOnlySeqLib
from ..overlap import OverlapSeqLib


SEQLIB_LABEL_TEXT = OrderedDict(
    [
        ("BcvSeqLib", "Barcoded Variant"),
        ("BcidSeqLib", "Barcoded Identifier"),
        ("OverlapSeqLib", "Overlap"),
        ("BasicSeqLib", "Basic"),
        ("BarcodeSeqLib", "Barcodes Only"),
        ("IdOnlySeqLib", "Identifiers Only"),
    ]
)

#: map class names to class definitions to avoid use of globals()
SEQLIB_CLASSES = {
    "BarcodeSeqLib": BarcodeSeqLib,
    "BcvSeqLib": BcvSeqLib,
    "BcidSeqLib": BcidSeqLib,
    "BasicSeqLib": BasicSeqLib,
    "IdOnlySeqLib": IdOnlySeqLib,
    "OverlapSeqLib": OverlapSeqLib,
}


class CreateSeqLibDialog(tkinter.simpledialog.Dialog):
    """
    Dialog box for creating a new SeqLib.
    """

    def __init__(self, parent_window, title="New SeqLib"):
        self.element_tkstring = tk.StringVar()
        self.element_type = None
        tkinter.simpledialog.Dialog.__init__(self, parent_window, title)

    def body(self, master):
        message = tkinter.ttk.Label(master, text="SeqLib type:")
        message.grid(column=0, row=0)

        for i, k in enumerate(SEQLIB_LABEL_TEXT.keys()):
            rb = tkinter.ttk.Radiobutton(
                master,
                text=SEQLIB_LABEL_TEXT[k],
                variable=self.element_tkstring,
                value=k,
            )
            rb.grid(column=0, row=(i + 1), sticky="w")
            if i == 0:
                rb.invoke()

    def buttonbox(self):
        """
        Display only one button.
        """
        box = tk.Frame(self)

        w = tk.Button(box, text="OK", width=10, command=self.ok, default="active")
        w.pack(side="left", padx=5, pady=5)

        self.bind("<Return>", self.ok)

        box.pack()

    def apply(self):
        try:
            self.element_type = SEQLIB_CLASSES[self.element_tkstring.get()]
        except KeyError:
            raise KeyError("Unrecognized element type.")




import pandas as pd
import logging

from matplotlib.backends.backend_pdf import PdfPages
import os.path
from .plots import overlap_merge_plot
from .seqlib import SeqLib
from .variant import VariantSeqLib
from fqfa import open_compressed, parse_fastq_pe_reads, has_fastq_ext
from fqfa.fastq.fastqread import FastqRead
from .fastqheader import fastq_read_is_chaste


class OverlapSeqLib(VariantSeqLib):
    """
    Class for count data from sequencing libraries with overlapping paired-end 
    reads for each variant. Creating a 
    :py:class:`~seqlib.overlap.OverlapSeqLib` requires a valid *config* object 
    with an ``'overlap'`` entry.

    The ``"fastq"`` config entry must contain two read files, with the keys 
    ``"forward"`` and ``"reverse"``. Information about how to combine these 
    reads is in the ``"overlap"`` config entry.

    The ``"overlap"`` config entry contains the following keys:

    * ``"forward start"`` --- position in the forward read where the \
        overlapping region begins
    * ``"reverse start"`` --- position in the reverse read where the \
        overlapping region begins (before being reverse-complemented)
    * ``"length"`` --- number of bases in the overlapping region
    * ``"max mismatches"`` --- maximum number of mismatches tolerated in the \
        overlapping region before discarding the read
    * ``"overlap only"`` --- whether to trim the merged read to contain only \
        the overlapping region (optional, defaults to ``False``)

    Here is a schematic of the case in the above JSON example::

        forward ---> 1   
                     CGACGCAAGGA
                       |||||||||
                       ACTCCTTGCGTCG
                                   1 <--- reverse

    Note that the merged sequence is identical to the wild type sequence given 
    in the JSON file.
    """

    treeview_class_name = "Overlap SeqLib"

    def __init__(self):
        VariantSeqLib.__init__(self)
        self.forward = None
        self.reverse = None
        self.fwd_start = None
        self.rev_start = None
        self.overlap_length = None
        self.trim = None
        self.max_overlap_mismatches = None
        self.merge_mismatches = None
        self.default_filters.update({"merge failure": True})
        self.default_filters.update({"remove unresolvable": False})
        self.logger = logging.getLogger("{}.{}".format(__name__, self.__class__))

    def configure(self, cfg):
        """
        Set up the object using the config object *cfg*, usually derived from 
        a ``.json`` file.
        """
        VariantSeqLib.configure(self, cfg)
        self.logger = logging.getLogger(
            "{}.{} - {}".format(__name__, self.__class__.__name__, self.name)
        )

        # if counts are specified, copy them later
        # else handle the FASTQ config options and check the files
        if self.counts_file is None:
            self.configure_fastq(cfg)
            try:
                self.fwd_start = int(cfg["overlap"]["forward start"])
                self.rev_start = int(cfg["overlap"]["reverse start"])
                self.overlap_length = int(cfg["overlap"]["length"])
                self.trim = cfg["overlap"]["trim"]
                self.max_overlap_mismatches = int(cfg["overlap"]["max mismatches"])

                forward_error = False
                reverse_error = False
                if not has_fastq_ext(self.forward):
                    forward_error = True
                if not has_fastq_ext(self.reverse):
                    reverse_error = True
                if forward_error and reverse_error:
                    raise IOError(
                        "FASTQ file error: unrecognized extension (forward and reverse) [{}]".format(
                            self.name
                        )
                    )
                elif forward_error:
                    raise IOError(
                        "FASTQ file error: unrecognized extension (forward) [{}]".format(
                            self.name
                        )
                    )
                elif reverse_error:
                    raise IOError(
                        "FASTQ file error: unrecognized extension (reverse) [{}]".format(
                            self.name
                        )
                    )
            except IOError as fqerr:
                raise IOError("FASTQ file error [{}]: {}".format(self.name, fqerr))
            except KeyError as key:
                raise KeyError(
                    "Missing required config value {key} [{name}]".format(
                        key=key, name=self.name
                    )
                )
            except ValueError as value:
                raise ValueError(
                    "Invalid parameter value {value} [{name}]".format(
                        value=value, name=self.name
                    )
                )

    def serialize(self):
        """
        Format this object (and its children) as a config object suitable for dumping to a config file.
        """
        cfg = VariantSeqLib.serialize(self)

        cfg["fastq"] = self.serialize_fastq()
        cfg["overlap"] = {
            "forward start": self.fwd_start,
            "reverse start": self.rev_start,
            "length": self.overlap_length,
            "trim": self.trim,
            "max mismatches": self.max_overlap_mismatches,
        }

        return cfg

    def configure_fastq(self, cfg):
        """
        Set up the object's FASTQ_ file handling and filtering options.
        """
        try:
            self.forward = cfg["fastq"]["forward reads"]
            self.reverse = cfg["fastq"]["reverse reads"]

            if "merge failure" in cfg["fastq"]["filters"]:
                raise ValueError(
                    "'merge failure' is not user-configurable [{}]".format(self.name)
                )
            self.filters = cfg["fastq"]["filters"]
        except KeyError as key:
            raise KeyError(
                "Missing required config value {key} [{name}]".format(
                    key=key, name=self.name
                )
            )

    def serialize_fastq(self):
        """
        Serialize this object's FASTQ_ file handling and filtering options.
        """
        fastq = {
            "forward reads": self.forward,
            "reverse reads": self.reverse,
            "filters": self.serialize_filters(),
        }

        return fastq

    def merge_reads(self, fwd, rev):
        """
        Combines the *fwd* and *rev* FASTQ read objects into a
        single FASTQ read with the same header information as
        *fwd*. Mismatches are resolved by taking the highest quality base. If 
        discrepant bases have the same quality value, this position is 
        unresolvable and an ``'X'`` is inserted. Quality values in the 
        resulting FASTQ read are the maximum quality for the
        given base at that position. Returns ``None`` if the maximum number of 
        mismatches in the overlap region is exceded.
        """
        rev.reverse_complement()

        # print(fwd.sequence, "-" * (self.rev_start - 1), sep="")
        # print("-" * (self.fwd_start - 1), rev.sequence, sep="")
        rev_extra_start = len(rev) - self.rev_start + 1
        fwd_end = self.fwd_start + self.overlap_length - 1
        merge = FastqRead(
            header=fwd.header + "|" + rev.header,
            sequence="A",
            header2=fwd.header2 + "|" + rev.header2,
            quality_string="#",
            quality_encoding_value=fwd.quality_encoding_value,
        )
        merge.sequence = fwd.sequence[:fwd_end] + rev.sequence[rev_extra_start:]
        merge.quality = fwd.quality[:fwd_end] + rev.quality[rev_extra_start:]

        mismatches = 0
        first = True
        for i in range(self.overlap_length):
            a = self.fwd_start - 1 + i
            b = len(rev) - self.rev_start - self.overlap_length + i + 1
            try:
                if fwd.sequence[a] == rev.sequence[b]:
                    # take the highest quality value
                    if rev.quality[b] > fwd.quality[a]:
                        merge.quality[a] = rev.quality[b]
                else:
                    if fwd.quality[a] == rev.quality[b]:
                        merge.sequence[a] = "X"  # unresolvable
                        self.merge_mismatches.iloc[i]["unresolved"] += 1
                    elif rev.quality[b] > fwd.quality[a]:
                        merge.sequence[a] = rev.sequence[b]
                        merge.quality[a] = rev.quality[b]
                        self.merge_mismatches.iloc[i]["resolved"] += 1
                    else:
                        # overlap region already same as fwd
                        self.merge_mismatches.iloc[i]["resolved"] += 1
                    mismatches += 1
                    if first:
                        self.merge_mismatches.iloc[i]["first"] += 1
                        first = False
            except IndexError:
                raise IndexError(
                    "Failed to calculate overlap (a={a}, len(a)={lena}, b={b}, len(b)={lenb}) [{name}]".format(
                        a=a,
                        b=b,
                        lena=len(fwd.sequence),
                        lenb=len(rev.sequence),
                        name=self.name,
                    )
                )

        if mismatches > self.max_overlap_mismatches:
            return None  # merge failed

        if self.trim:
            merge.trim_length(self.overlap_length, self.fwd_start)
        return merge

    def counts_from_reads(self):
        df_dict = dict()

        self.merge_mismatches = pd.DataFrame(
            data=0,
            index=[
                x + self.fwd_start + self.wt.dna_offset
                for x in range(0, self.overlap_length)
            ],
            columns=["resolved", "unresolved", "first"],
        )

        self.logger.info("Counting variants")
        max_mut_variants = 0
        with open_compressed(self.forward) as handle_fwd, open_compressed(self.reverse) as handle_rev:
            for fwd, rev in parse_fastq_pe_reads(handle_fwd, handle_rev):
                # filter on chastity before merge
                chaste = True
                if self.filters["chastity"]:
                    if not fastq_read_is_chaste(fwd):
                        chaste = False
                        if self.report_filtered:
                            self.report_filtered_read(fwd, "chastity")
                    if not fastq_read_is_chaste(rev):
                        chaste = False
                        if self.report_filtered:
                            self.report_filtered_read(rev, "chastity")
                    if not chaste:
                        self.filter_stats["chastity"] += 1
                        self.filter_stats["total"] += 1
                        continue

                merge = self.merge_reads(fwd, rev)
                if merge is None:  # merge failed
                    self.filter_stats["merge failure"] += 1
                    self.filter_stats["total"] += 1
                    if self.report_filtered:
                        self.report_filtered_read(fwd, {"merge failure": True})
                        self.report_filtered_read(rev, {"merge failure": True})
                else:
                    if self.read_quality_filter(merge):
                        mutations = self.count_variant(merge.sequence)
                        if mutations is None:  # merge read has too many mutations
                            max_mut_variants += 1
                            if self.report_filtered:
                                self.report_filtered_variant(merge.sequence, 1)
                        else:
                            try:
                                df_dict[mutations] += 1
                            except KeyError:
                                df_dict[mutations] = 1

        self.store.put(
            "/raw/overlap_mismatches",
            self.merge_mismatches,
            format="table",
            data_columns=self.merge_mismatches.columns,
        )
        self.merge_mismatches = None
        self.save_counts("variants", df_dict, raw=True)
        del df_dict

        if self.aligner is not None:
            self.logger.info("Aligned {} variants".format(self.aligner.calls))
            self.aligner_cache = None
        self.logger.info(
            "Removed {} total variants with excess mutations"
            "".format(max_mut_variants)
        )
        self.save_filter_stats()

    def calculate(self):
        """
        Reads the forward and reverse reads, merges them, performs 
        quality-based filtering, and counts the variants.
        """
        if not self.check_store("/main/variants/counts"):
            if not self.check_store("/raw/variants/counts"):
                if self.counts_file is not None:
                    self.counts_from_file(self.counts_file)
                else:  # count everything
                    self.counts_from_reads()
            self.save_filtered_counts("variants", "count >= self.variant_min_count")

        self.count_synonymous()

    def make_plots(self):
        """
        Make plots for :py:class:`~seqlib.seqlib.OverlapSeqLib` objects.

        Creates plots of the location of merged read mismatches.
        """
        if self.plots_requested:
            SeqLib.make_plots(self)
            pdf = PdfPages(os.path.join(self.plot_dir, "overlap_mismatches.pdf"))
            overlap_merge_plot(self, pdf)
            pdf.close()




import tkinter as tk
import tkinter.ttk
import tkinter.simpledialog
from .dialog_elements import FileEntry, StringEntry, DEFAULT_COLUMNS
from .create_seqlib_dialog import SEQLIB_LABEL_TEXT
from ..barcode import BarcodeSeqLib
from ..barcodevariant import BcvSeqLib
from ..barcodeid import BcidSeqLib
from ..basic import BasicSeqLib
from ..idonly import IdOnlySeqLib
from ..overlap import OverlapSeqLib
from ..selection import Selection
from ..experiment import Experiment


#: map class names to class definitions to avoid use of globals()
ELEMENT_CLASSES = {
    "BarcodeSeqLib": BarcodeSeqLib,
    "BcvSeqLib": BcvSeqLib,
    "BcidSeqLib": BcidSeqLib,
    "BasicSeqLib": BasicSeqLib,
    "IdOnlySeqLib": IdOnlySeqLib,
    "OverlapSeqLib": OverlapSeqLib,
    "Selection": Selection,
    "Experiment": Experiment,
}


class CreateRootDialog(tkinter.simpledialog.Dialog):
    """
    Dialog box for creating a new root element.
    """

    def __init__(self, parent_window, title="Create Root Object"):
        self.element_tkstring = tk.StringVar()
        self.cfg_dict = dict()
        self.output_directory_tk = FileEntry(
            "Output Directory",
            self.cfg_dict,
            "output directory",
            optional=False,
            directory=True,
        )
        self.name_tk = StringEntry("Name", self.cfg_dict, "name", optional=False)
        self.element = None
        tkinter.simpledialog.Dialog.__init__(self, parent_window, title)

    def body(self, master):
        row_no = self.name_tk.body(master, 0)
        row_no += self.output_directory_tk.body(master, row_no)

        element_types = tkinter.ttk.Frame(master, padding=(3, 3, 12, 12))
        element_types.grid(
            column=0, row=row_no, sticky="nsew", columnspan=DEFAULT_COLUMNS
        )

        message = tkinter.ttk.Label(element_types, text="Root object type:")
        message.grid(column=0, row=0)

        label = tkinter.ttk.Label(element_types, text="Experiment")
        label.grid(column=0, row=1, sticky="w")
        rb = tkinter.ttk.Radiobutton(
            element_types,
            text="Experiment",
            variable=self.element_tkstring,
            value="Experiment",
        )
        rb.grid(column=0, row=2, sticky="w")
        rb.invoke()

        label = tkinter.ttk.Label(element_types, text="Selection")
        label.grid(column=0, row=3, sticky="w")
        rb = tkinter.ttk.Radiobutton(
            element_types,
            text="Selection",
            variable=self.element_tkstring,
            value="Selection",
        )
        rb.grid(column=0, row=4, sticky="w")

        label = tkinter.ttk.Label(element_types, text="SeqLib")
        label.grid(column=0, row=5, sticky="w")
        for i, k in enumerate(SEQLIB_LABEL_TEXT.keys()):
            rb = tkinter.ttk.Radiobutton(
                element_types,
                text=SEQLIB_LABEL_TEXT[k],
                variable=self.element_tkstring,
                value=k,
            )
            rb.grid(column=0, row=(i + 6), sticky="w")

    def buttonbox(self):
        """
        Display only one button.
        """
        box = tk.Frame(self)

        w = tk.Button(box, text="OK", width=10, command=self.ok, default="active")
        w.pack(side="left", padx=5, pady=5)

        self.bind("<Return>", self.ok)

        box.pack()

    def validate(self):
        # check the fields
        return self.output_directory_tk.validate() and self.name_tk.validate()

    def apply(self):
        # apply the fields
        self.output_directory_tk.apply()
        self.name_tk.apply()

        # create the object
        try:
            self.element = ELEMENT_CLASSES[self.element_tkstring.get()]()
        except KeyError:
            raise KeyError(
                "Unrecognized element type '{}'".format(self.element_tkstring.get())
            )

        # set the properties from this dialog
        self.element.output_dir_override = False
        self.element.output_dir = self.cfg_dict["output directory"]
        self.element.name = self.cfg_dict["name"]



import re

# Matches FASTQ headers based on the following pattern (modify as needed):
# @<instrument>:<run number>:<flowcell ID>:<lane>:<tile>:<x-pos>:<y-pos> <read>:<is filtered>:<control number>:<sample number>

# Example: @M02564:876:000000000-L3775:1:1101:16862:1800 1:N:0:TCACTCGA+TAACGGTT
# Sample number contains indexes if they are present.

# See: https://help.basespace.illumina.com/files-used-by-basespace/fastq-files

# Note: this regex is currently unused since it is not needed to support the legacy chastity filtering feature
new_header_pattern = re.compile(
    r"""
    @(?P<Instrument>[^:]+):
    (?P<RunNumber>\d+):
    (?P<FlowcellID>[^:]+):
    (?P<Lane>\d+):
    (?P<Tile>\d+):
    (?P<XPos>\d+):
    (?P<YPos>\d+)
    \s
    (?P<Read>\d+):
    (?P<IsFiltered>[YN]):
    (?P<ControlNumber>[^:]+):
    (?P<SampleNumber>[^:]+)
    """,
    re.VERBOSE,
)

# Matches FASTQ headers based on the following pattern (modify as needed):
# @<MachineName>:<Lane>:<Tile>:<X>:<Y>:<Chastity>#<IndexRead>/<ReadNumber>
old_header_pattern = re.compile(
    r"""
    @(?P<MachineName>.+):
    (?P<Lane>\d+):
    (?P<Tile>\d+):
    (?P<X>\d+):
    (?P<Y>\d+):
    (?P<Chastity>[01])#
    (?P<IndexRead>\d)/
    (?P<ReadNumber>\d)
    """,
    re.VERBOSE,
)

def parse_fastq_header(fq, pattern=old_header_pattern):
    """Parse the read's FASTQ_ header and return key-value pairs.

    Parses the first FASTQ_ header (@ header) and returns a dictionary.
    Dictionary keys are the named groups in the regular expression
    *pattern*. Unnamed matches are ignored. Integer values are converted
    from strings to integers.

    The default pattern matches a header in the format::

        @<MachineName>:<Lane>:<Tile>:<X>:<Y>:<Chastity>#<IndexRead>/<ReadNumber>

    """
    match = pattern.match(fq.header)
    if match is None:
        return None
    else:
        header_dict = match.groupdict()
        for key in header_dict:
            if header_dict[key].isdigit():
                header_dict[key] = int(header_dict[key])
        return header_dict


def fastq_read_is_chaste(self, raises=True):
    """
    Returns ``True`` if the chastity bit is set in the header. The
    regular experession used by :py:meth:`header_information` must
    include a ``'Chastity'`` match that equals ``1`` if the read is
    chaste.

    If ``raises`` is ``True``, raises an informative error if the
    chastity information in the header is not found. Otherwise, a
    read without chastity information is treated as unchaste.
    """
    try:
        if self.header_information()["Chastity"] == 1:
            return True
        else:
            return False
    except KeyError:  # no 'Chastity' in pattern
        if raises:
            raise KeyError("No chastity bit in FASTQ header pattern")
        else:
            return False
    except TypeError:  # no header match (unexpected format)
        if raises:
            raise ValueError("Unexpected FASTQ header format")
        else:
            return False




import re
from .aligner import Aligner
from .seqlib import SeqLib
from .wildtype import WildTypeSequence
import logging
from .constants import WILD_TYPE_VARIANT, SYNONYMOUS_VARIANT
from .constants import CODON_TABLE, AA_CODES

#: Default number of maximum mutation.
#: Must be set to avoid data frame performance errors.
DEFAULT_MAX_MUTATIONS = 10

#: Matches a single amino acid substitution in HGVS_ format.
re_protein = re.compile(
    r"(?P<match>p\.(?P<pre>[A-Z][a-z][a-z])(?P<pos>-?\d+)" "(?P<post>[A-Z][a-z][a-z]))"
)

#: Matches a single nucleotide substitution (coding or noncoding)
#: in HGVS_ format.
re_nucleotide = re.compile(
    r"(?P<match>[nc]\.(?P<pos>-?\d+)(?P<pre>[ACGT])>(?P<post>[ACGT]))"
)

#: Matches a single coding nucleotide substitution in HGVS_ format.
re_coding = re.compile(
    r"(?P<match>c\.(?P<pos>-?\d+)(?P<pre>[ACGT])>(?P<post>[ACGT]) \(p\.(?:=|[A-Z][a-z][a-z]-?\d+[A-Z][a-z][a-z])\))"
)

#: Matches a single noncoding nucleotide substitution in HGVS_ format.
re_noncoding = re.compile(
    r"(?P<match>n\.(?P<pos>-?\d+)(?P<pre>[ACGT])>(?P<post>[ACGT]))"
)


def valid_variant(s, is_coding=True):
    """
    Returns True if s is a valid coding or noncoding variant, else False.
    """
    if s == WILD_TYPE_VARIANT:
        return True
    else:
        if is_coding:
            for mut in s.split(", "):
                match = re_coding.match(mut)
                if match is None:
                    return False
            return True
        else:
            for mut in s.split(", "):
                match = re_noncoding.match(mut)
                if match is None:
                    return False
            return True


def hgvs2single(s):
    """
    Convert HGVS string from Enrich2 to <pre><pos><post>
    single-letter amino acid variant string.

    Returns a list of single-letter variants.
    """
    t = re_protein.findall(s)
    return ["{}{}{}".format(AA_CODES[m[1]], m[2], AA_CODES[m[3]]) for m in t]


def single2hgvs(s):
    """
    Convert single-letter amino acid changes in the form
    <pre><pos><post> into HGVS strings that match Enrich2
    output.

    Searches the string s for all instances of the above
    pattern and returns a list of Enrich2 variants.
    """
    t = re.findall(r"[A-Z*]\d+[A-Z*]", s)
    return ["p.{}{}{}".format(AA_CODES[x[0]], x[1:-1], AA_CODES[x[-1]]) for x in t]


def get_variant_type(variant):
    """
    Use regular expressions to determine whether the variant is protein,
    coding, or noncoding.

    :param str variant: variant string
    :return: ``'protein'``, ``'coding'``, or ``'noncoding'`` depending on \
    which regular expression matches, else ``None``. Note that both wild type \
    and synonymous special variants will return ``None``.
    :rtype: str
    """
    v = variant.split(", ")[0]  # test first token of multi-mutant
    if re_protein.match(v) is not None:
        return "protein"
    elif re_coding.match(v) is not None:
        return "coding"
    elif re_noncoding.match(v) is not None:
        return "noncoding"
    else:
        return None


def mutation_count(variant):
    """
    Counts the number of mutations in the HGVS_ *variant*.

    :param str variant: variant string
    :return: number of variants (0 if wild type or synonymous)
    :rtype: int
    """
    if len(variant) == 0:
        raise ValueError("Empty variant string.")
    elif variant == WILD_TYPE_VARIANT:
        return 0
    elif variant == SYNONYMOUS_VARIANT:
        return 0
    else:
        return len(variant.split(","))


def has_indel(variant):
    """
    Tests if the HGVS_ *variant* contains an indel mutation.

    :param str variant: variant string
    :return: ``True`` if there is an indel, else ``False``
    :rtype: bool
    """
    if len(variant) == 0:
        raise ValueError("Empty variant string.")
    elif variant == WILD_TYPE_VARIANT:
        return False
    elif variant == SYNONYMOUS_VARIANT:
        return False
    else:
        return any(x in variant for x in ("ins", "del", "dup"))


def has_unresolvable(variant):
    """
    Tests if the HGVS_ *variant* has an unresolvable amino acid change.

    Unresolvable amino acid changes are most commonly caused by the presence
    of N or X nucleotides, resulting in a non-translatable codon. They are
    also found when a frameshift causes the last part of the coding sequence
    to not have three nucleotides.

    :param str variant: variant string
    :return: ``True`` if there is an unresolvable change, else ``False``
    :rtype: bool
    """
    if AA_CODES["?"] in variant:
        return True
    else:
        return False


def protein_variant(variant):
    """
    Return an HGVS_ variant string containing only the protein changes in a
    coding HGVS_ variant string. If all variants are synonymous, returns the
    synonymous variant code. If the variant is wild type, returns the wild
    type variant.

    :param str variant: coding variant string
    :return: protein variant string (or synonymous or wild type)
    :rtype: str
    """
    if len(variant) == 0:
        raise ValueError("Empty variant string.")
    elif variant == WILD_TYPE_VARIANT:
        return WILD_TYPE_VARIANT
    elif variant == SYNONYMOUS_VARIANT:
        return SYNONYMOUS_VARIANT
    else:
        matches = re.findall(r"\((p\.\S*)\)", variant)
        if len(matches) == 0:
            raise ValueError("Invalid coding variant string.")
        # uniqify and remove synonymous
        seen = {"p.=": True}
        unique_matches = list()
        for v in matches:
            if v in seen:
                continue
            else:
                seen[v] = True
                unique_matches.append(v)
        if len(unique_matches) == 0:
            return SYNONYMOUS_VARIANT
        else:
            return ", ".join(unique_matches)


class VariantSeqLib(SeqLib):
    """
    Abstract :py:class:`~seqlib.seqlib.SeqLib` class for for Enrich libraries
    containing variants. Implements core functionality for assessing variants,
    either coding or noncoding. Subclasess must evaluate the variant DNA
    sequences that are being counted.
    """

    def __init__(self):
        SeqLib.__init__(self)
        self.wt = WildTypeSequence(self.name)
        self.aligner = None
        self.aligner_cache = None
        self.variant_min_count = 0
        self.max_mutations = DEFAULT_MAX_MUTATIONS
        # 'synonymous' label may be added in configure() if wt is coding
        self.add_label("variants")
        self.logger = logging.getLogger("{}.{}".format(__name__, self.__class__))

    def configure(self, cfg):
        """
        Set up the object using the config object *cfg*, usually derived from
        a ``.json`` file.
        """
        SeqLib.configure(self, cfg)
        self.logger = logging.getLogger(
            "{}.{} - {}".format(__name__, self.__class__.__name__, self.name)
        )

        self.wt.configure(cfg["variants"]["wild type"])
        if self.is_coding():
            self.add_label("synonymous")
        try:
            if "use aligner" in cfg["variants"]:
                if cfg["variants"]["use aligner"]:
                    self.aligner = Aligner()
                    self.aligner_cache = dict()
                else:
                    self.aligner = None
                    self.aligner_cache = None

            if "min count" in cfg["variants"]:
                self.variant_min_count = int(cfg["variants"]["min count"])

            if "max mutations" in cfg["variants"]:
                self.max_mutations = int(cfg["variants"]["max mutations"])

        except KeyError as key:
            raise KeyError(
                "Missing required config value {key} [{name}]"
                "".format(key=key, name=self.name)
            )

    def serialize(self):
        """
        Format this object (and its children) as a config object suitable for
        dumping to a config file.
        """
        cfg = SeqLib.serialize(self)

        cfg["variants"] = dict()
        cfg["variants"]["wild type"] = self.wt.serialize()
        cfg["variants"]["use aligner"] = self.aligner is not None
        if self.max_mutations != DEFAULT_MAX_MUTATIONS:
            cfg["variants"]["max mutations"] = self.max_mutations
        if self.variant_min_count > 0:
            cfg["variants"]["min count"] = self.variant_min_count

        return cfg

    def is_coding(self):
        """
        Return ``True`` if the variants are protein-coding, else ``False``.
        """
        return self.wt.is_coding()

    def has_wt_sequence(self):
        """
        Returns ``True``, because :py:class:`~seqlib.seqlib.VariantSeqLib`
        objects have a wild type sequence. Raises a ValueError if the wild type
        sequence is not set properly.
        """
        if self.wt is not None:
            return True
        else:
            raise ValueError("Wild type not set properly [{}]".format(self.name))

    def align_variant(self, variant_dna):
        """
        Use the local :py:class:`~seqlib.aligner.Aligner` instance to align the
        *variant_dna* to the wild type sequence. Returns a list of HGVS_
        variant strings.

        Aligned variants are stored in a local dictionary to avoid recomputing
        alignments. This dictionary should be cleared after all variants are
        counted, to save memory.

        .. warning:: Using the :py:class:`~seqlib.aligner.Aligner` \
        dramatically increases runtime.
        """
        if variant_dna in list(self.aligner_cache.keys()):
            return self.aligner_cache[variant_dna]

        mutations = list()
        traceback = self.aligner.align(self.wt.dna_seq, variant_dna)
        for x, y, cat, length in traceback:
            if cat == "match":
                continue
            elif cat == "mismatch":
                mut = "{pre}>{post}".format(pre=self.wt.dna_seq[x], post=variant_dna[y])
            elif cat == "insertion":
                if y > length:
                    dup = variant_dna[y : y + length]
                    if dup == variant_dna[y - length : y]:
                        mut = "dup{seq}".format(seq=dup)
                    else:
                        mut = "_{pos}ins{seq}".format(pos=x + 2, seq=dup)
                else:
                    mut = "_{pos}ins{seq}".format(
                        pos=x + 2, seq=variant_dna[y : y + length]
                    )
            elif cat == "deletion":
                mut = "_{pos}del".format(pos=x + length)
            else:
                raise ValueError("Unable to resolve mutation [{}]".format(self.name))
            mutations.append((x, mut))

        self.aligner_cache[variant_dna] = mutations
        return mutations

    def count_variant(self, variant_dna, include_indels=True):
        """
        Identifies mutations and counts the *variant_dna* sequence.
        The algorithm attempts to call variants by comparing base-by-base.
        If the *variant_dna* and wild type DNA are different lengths, or if
        there are an excess of mismatches (indicating a possible indel), local
        alignment is performed using :py:meth:`align_variant` if this option
        has been selected in the configuration.

        Each variant is stored as a tab-delimited string of mutations in HGVS_
        format. Returns a list of HGVS_ variant strings. Returns an empty list
        if the variant is wild type. Returns None if the variant was discarded
        due to excess mismatches.
        """
        if not re.match("^[ACGTNXacgtnx]+$", variant_dna):
            raise ValueError(
                "Variant DNA sequence contains unexpected "
                "characters [{}]".format(self.name)
            )

        variant_dna = variant_dna.upper()

        if len(variant_dna) != len(self.wt.dna_seq):
            if self.aligner is not None:
                mutations = self.align_variant(variant_dna)
            else:
                return None
        else:
            mutations = list()
            for i in range(len(variant_dna)):
                if variant_dna[i] != self.wt.dna_seq[i]:
                    mutations.append(
                        (
                            i,
                            "{pre}>{post}".format(
                                pre=self.wt.dna_seq[i], post=variant_dna[i]
                            ),
                        )
                    )
                    if len(mutations) > self.max_mutations:
                        if self.aligner is not None:
                            mutations = self.align_variant(variant_dna)
                            if len(mutations) > self.max_mutations:
                                # too many mutations post-alignment
                                return None
                            else:
                                # stop looping over this variant
                                break
                        else:
                            # too many mutations and not using aligner
                            return None

        mutation_strings = list()
        if self.is_coding():
            variant_protein = ""
            for i in range(0, len(variant_dna), 3):
                try:
                    variant_protein += CODON_TABLE[variant_dna[i : i + 3]]
                except KeyError:  # garbage codon due to indel, X, or N
                    variant_protein += "?"

            for pos, change in mutations:
                ref_dna_pos = pos + self.wt.dna_offset + 1
                ref_pro_pos = int(pos / 3) + self.wt.protein_offset + 1
                mut = "c.{pos}{change}".format(pos=ref_dna_pos, change=change)
                if has_indel(change):
                    mut += " (p.{pre}{pos}fs)".format(
                        pre=AA_CODES[self.wt.protein_seq[int(pos / 3)]], pos=ref_pro_pos
                    )
                elif variant_protein[int(pos / 3)] == self.wt.protein_seq[int(pos / 3)]:
                    mut += " (p.=)"
                else:
                    mut += " (p.{pre}{pos}{post})".format(
                        pre=AA_CODES[self.wt.protein_seq[int(pos / 3)]],
                        pos=ref_pro_pos,
                        post=AA_CODES[variant_protein[int(pos / 3)]],
                    )
                mutation_strings.append(mut)
        else:
            for pos, change in mutations:
                ref_dna_pos = pos + self.wt.dna_offset + 1
                mut = "n.{pos}{change}".format(pos=ref_dna_pos, change=change)
                mutation_strings.append(mut)

        if len(mutation_strings) > 0:
            variant_string = ", ".join(mutation_strings)
        else:
            variant_string = WILD_TYPE_VARIANT
        return variant_string

    def count_synonymous(self):
        """
        Combine counts for synonymous variants (defined as variants that differ
        at the nucleotide level but not at the amino acid level) and store them
        under the ``synonymous`` label.

        This method should be called only after ``variants`` have been counted.

        .. note:: The total number of ``synonymous`` variants may be greater \
        than the total number of ``variants`` after filtering. This is \
        because ``variants`` are combined into ``synonymous`` entries at the \
        :py:class:`~seqlib.seqlib.SeqLib` level before count-based filtering, \
        allowing filtered ``variants`` to contribute counts to their \
        ``synonymous`` entry.
        """
        if not self.is_coding():
            self.logger.warning("Cannot count synonymous mutations in noncoding data")
            return

        if self.check_store("/main/synonymous/counts"):
            return

        self.logger.info("Counting synonymous variants")

        df_dict = dict()

        for variant, count in self.store["/main/variants/counts"].iterrows():
            if variant == WILD_TYPE_VARIANT:
                df_dict[variant] = count["count"]
            else:
                variant = protein_variant(variant)
                if len(variant) == 0:
                    variant = SYNONYMOUS_VARIANT
                try:
                    df_dict[variant] += count["count"]
                except KeyError:
                    df_dict[variant] = count["count"]

        self.save_counts("synonymous", df_dict, raw=False)
        del df_dict

    def report_filtered_variant(self, variant, count):
        """
        Outputs a summary of the filtered variant to *handle*. Internal filter
        names are converted to messages using the ``SeqLib.filter_messages``
        dictionary. Related to :py:meth:`SeqLib.report_filtered`.
        """
        self.logger.debug(
            "Filtered variant (quantity={n}) (excess mutations)"
            "\n{read!s}".format(n=count, read=variant)
        )




import logging
import os.path
import pandas as pd
import numpy as np
from collections import OrderedDict
from matplotlib.backends.backend_pdf import PdfPages
import sys
from .plots import counts_plot
from .storemanager import StoreManager, fix_filename, ELEMENT_LABELS
from .fastqheader import fastq_read_is_chaste


class SeqLib(StoreManager):
    """
    Abstract class for handling count data from a single sequencing library.
    """

    # Note: the following block is referenced by line number above
    # When adding new messages, update the documentation line numbers also!
    filter_messages = OrderedDict(
        [
            ("min quality", "single-base quality"),
            ("avg quality", "average quality"),
            ("max N", "excess N bases"),
            ("chastity", "not chaste"),
            ("remove unresolvable", "unresolvable mismatch"),
            ("merge failure", "unable to merge reads"),
            ("total", "total"),
        ]
    )

    store_suffix = "lib"

    def __init__(self):
        StoreManager.__init__(self)
        self.logger = logging.getLogger("{}.{}".format(__name__, self.__class__))
        self.timepoint = None
        self.counts_file = None
        self.report_filtered = None
        self._filters = dict()
        self.filter_stats = dict()
        self.default_filters = dict()
        self.default_filters.update({"min quality": 0})
        self.default_filters.update({"max N": sys.maxsize})
        self.default_filters.update({"avg quality": 0})
        self.default_filters.update({"chastity": False})

    @property
    def filters(self):
        return self._filters

    @filters.setter
    def filters(self, config_filters):
        """
        Set up the filters dictionary using the options selected in
        *config_filters*, filling in missing entries with defaults.
        """
        self._filters.clear()
        self._filters.update(self.default_filters)

        unused = list()
        for key in config_filters:
            if key in self._filters:
                if config_filters[key] is not None:
                    self._filters[key] = config_filters[key]
            else:
                unused.append(key)
        if len(unused) > 0:
            self.logger.warning(
                "Unused filter parameters ({})" "".format(", ".join(unused))
            )

        self.filter_stats.clear()
        for key in self._filters:
            self.filter_stats[key] = 0
        self.filter_stats["total"] = 0

    def serialize_filters(self):
        """
        Return a dictionary of filtering options that have non-default values.
        """
        cfg = dict()
        for key in list(self.filters.keys()):
            if self.filters[key] != self.default_filters[key]:
                cfg[key] = self.filters[key]
        return cfg

    def _children(self):
        """
        These objects have no children. Returns ``None``.
        """
        return None

    def add_child(self, child):
        """
        No children, raises an AttributeError.
        """
        raise AttributeError("SeqLib objects do not support adding children")

    def remove_child_id(self, tree_id):
        """
        No children, raises an AttributeError.
        """
        raise AttributeError("SeqLib objects do not support removing children")

    def validate(self):
        """
        Validates paramaters for a configured SeqLib. Currently does nothing.
        """
        pass

    def has_wt_sequence(self):
        """
        Returns whether or not the object has a wild type sequence. Returns
        ``False`` unless overloaded by a derived class (such as
        :py:class:`~seqlib.seqlib.VariantSeqLib`).
        """
        return False

    def configure(self, cfg):
        """
        Set up the object using the config object *cfg*, usually derived from
        a ``.json`` file.
        """
        StoreManager.configure(self, cfg)
        self.logger = logging.getLogger(
            "{}.{} - {}".format(__name__, self.__class__.__name__, self.name)
        )
        try:
            self.timepoint = int(cfg["timepoint"])
            if "report filtered reads" in cfg:
                self.report_filtered = cfg["report filtered reads"]
            else:
                self.report_filtered = False
            if "counts file" in cfg:
                self.counts_file = cfg["counts file"]
            else:
                self.counts_file = None
        except KeyError as key:
            raise KeyError(
                "Missing required config value {key}" "".format(key=key), self.name
            )
        except ValueError as value:
            raise ValueError(
                "Invalid parameter value {value}" "".format(value=value), self.name
            )

    def serialize(self):
        """
        Format this object (and its children) as a config object suitable for
        dumping to a config file.
        """
        cfg = StoreManager.serialize(self)
        cfg["timepoint"] = self.timepoint
        cfg["report filtered reads"] = self.report_filtered
        if self.counts_file is not None:
            cfg["counts file"] = self.counts_file

        return cfg

    def calculate(self):
        """
        Pure virtual method that defines how the data are counted.
        """
        raise NotImplementedError("must be implemented by subclass")

    def report_filtered_read(self, fq, filter_flags):
        """
        Write the FASTQ read object *fq* to the ``DEBUG``
        log. The dictionary *filter_flags* contains ``True``
        values for each filtering option that applies to *fq*. Keys in
        *filter_flags* are converted to messages using the
        ``SeqLib.filter_messages`` dictionary.
        """
        self.logger.debug(
            "Filtered read ({messages})\n{read!s}\n{name}".format(
                messages=", ".join(
                    SeqLib.filter_messages[x] for x in filter_flags if filter_flags[x]
                ),
                name=self.name,
                read=fq,
            )
        )

    def save_counts(self, label, df_dict, raw):
        """
        Convert the counts in the dictionary *df_dict* into a DataFrame object
        and save it to the data store.

        If *raw* is ``True``, the counts are stored under
        ``"/raw/label/counts"``; else ``"/main/label/counts"``.
        """
        # Check if df_dict sequences are bytes, and convert to utf-8 if so
        if isinstance(list(df_dict.keys())[0], bytes):
            df_dict = {s.decode('utf-8') : v for (v,s) in enumerate(df_dict)}
        if len(list(df_dict.keys())) == 0:
            raise ValueError("Failed to count {} [{}]".format(label, self.name))
        df = pd.DataFrame.from_dict(df_dict, orient="index", dtype=np.int32)
        df.columns = ["count"]
        df.sort_values("count", ascending=False, inplace=True)
        self.logger.info(
            "Counted {n} {label} ({u} unique)".format(
                n=df["count"].sum(), u=len(df.index), label=label
            )
        )
        if raw:
            key = "/raw/{}/counts".format(label)
        else:
            key = "/main/{}/counts".format(label)
        self.store.put(key, df, format="table", data_columns=df.columns)
        del df

    def save_filtered_counts(self, label, query):
        """
        Filter the counts in ``"/raw/label/counts"`` using the *query* string
        and store the result in ``"/main/label/counts"``

        For more information on building query strings, see
        http://pandas.pydata.org/pandas-docs/stable/io.html#querying-a-table
        """
        self.logger.info("Converting raw {} counts to main counts".format(label))
        raw_table = "/raw/{}/counts".format(label)
        main_table = "/main/{}/counts".format(label)
        self.map_table(source=raw_table, destination=main_table, source_query=query)
        self.logger.info(
            "Counted {n} {label} ({u} unique) after query".format(
                n=self.store[main_table]["count"].sum(),
                u=len(self.store[main_table].index),
                label=label,
            )
        )

    def report_filter_stats(self):
        """
        Create report file for the number of filtered reads.

        The report file is located in the output directory, named
        ``SeqLibName.filter.txt``.
        It contains the number of reads filtered for each category, plus the
        total number filtered.

        .. note:: Reads are checked for all quality-based criteria before \
        filtering.
        """
        with open(
            os.path.join(self.output_dir, fix_filename(self.name) + ".filter.txt"), "w"
        ) as handle:
            for key in sorted(
                self.filter_stats, key=self.filter_stats.__getitem__, reverse=True
            ):
                if key != "total" and self.filter_stats[key] > 0:
                    print(
                        SeqLib.filter_messages[key],
                        self.filter_stats[key],
                        sep="\t",
                        file=handle,
                    )
            print("total", self.filter_stats["total"], sep="\t", file=handle)
        self.logger.info("Wrote filtering statistics")

    def save_filter_stats(self):
        """
        Save a DataFrame containing the number of filtered reads under
        ``'/raw/filter'``.

        This DataFrame contains the same information as ``report_filter_stats``
        """
        df = pd.DataFrame(index=list(SeqLib.filter_messages.values()), columns=["count"])
        for key in list(self.filter_stats.keys()):
            if self.filter_stats[key] > 0 or key == "total":
                df.loc[SeqLib.filter_messages[key], "count"] = self.filter_stats[key]
        df.dropna(inplace=True)
        self.store.put(
            "/raw/filter", df.astype(int), format="table", data_columns=df.columns
        )

    def read_quality_filter(self, fq):
        """
        Check the quality of the FastqRead object *fq*.

        Checks ``'chastity'``, ``'min quality'``, ``'avg quality'``,
        ``'max N'``, and ``'remove unresolvable'``.
        Counts failed reads for later output and reports the filtered read if
        desired.
        Returns ``True`` if the read passes all filters, else ``False``.
        """
        filter_flags = dict()
        for key in self.filters:
            filter_flags[key] = False

        if self.filters["chastity"]:
            if not fastq_read_is_chaste(fq):
                self.filter_stats["chastity"] += 1
                filter_flags["chastity"] = True

        if self.filters["min quality"] > 0:
            if fq.min_quality() < self.filters["min quality"]:
                self.filter_stats["min quality"] += 1
                filter_flags["min quality"] = True

        if self.filters["avg quality"] > 0:
            if fq.average_quality() < self.filters["avg quality"]:
                self.filter_stats["avg quality"] += 1
                filter_flags["avg quality"] = True

        if self.filters["max N"] >= 0:
            if fq.sequence.upper().count("N") > self.filters["max N"]:
                self.filter_stats["max N"] += 1
                filter_flags["max N"] = True

        if "remove unresolvable" in self.filters:  # OverlapSeqLib only
            if self.filters["remove unresolvable"]:
                if "X" in fq.sequence:
                    self.filter_stats["remove unresolvable"] += 1
                    filter_flags["remove unresolvable"] = True

        # update total and report if failed
        if any(filter_flags.values()):
            self.filter_stats["total"] += 1
            if self.report_filtered:
                self.report_filtered_read(fq, filter_flags)
            return False
        else:
            return True

    def make_plots(self):
        """
        Make plots that are shared by all :py:class:`~seqlib.seqlib.SeqLib`
        objects.

        Creates counts histograms for all labels.
        """
        if self.plots_requested:
            self.logger.info("Creating plots")

            pdf = PdfPages(os.path.join(self.plot_dir, "counts.pdf"))
            for label in self.labels:
                counts_plot(self, label, pdf, log=True)
                counts_plot(self, label, pdf, log=False)
            pdf.close()

    def write_tsv(self):
        """
        Write each table from the store to its own tab-separated file.

        Files are written to a ``tsv`` directory in the default output
        location.
        File names are the HDF5 key with ``'_'`` substituted for ``'/'``.
        """
        if self.tsv_requested:
            self.logger.info("Generating tab-separated output files")
            for k in list(self.store.keys()):
                self.write_table_tsv(k)

    def counts_from_file_h5(self, fname):
        """
        If an HDF store containing raw counts has been specified, open the
        store, copy those counts into this store, and close the counts store.

        Copies all tables in the ``'/raw'`` group along with their metadata.
        """
        store = pd.HDFStore(fname)
        self.logger.info(
            "Using existing HDF5 data store '{}' for raw data" "".format(fname)
        )
        # this could probably be much more efficient, but the PyTables docs
        # don't explain copying subsets of files adequately
        raw_keys = [key for key in list(store.keys()) if key.startswith("/raw/")]
        if len(raw_keys) == 0:
            raise ValueError(
                "No raw counts found in '{}' [{}]" "".format(fname, self.name)
            )
        else:
            for k in raw_keys:
                # copy the data table
                raw = store[k]
                self.store.put(k, raw, format="table", data_columns=raw.columns)
                # copy the metadata
                self.set_metadata(k, self.get_metadata(k, store=store), update=False)
                self.logger.info("Copied raw data '{}'".format(k))
        store.close()

    def counts_from_file_tsv(self, fname):
        """
        If a counts file in tsv format has been specified, read the counts into
        a new dataframe and save as raw counts.
        """
        df = pd.read_table(fname, sep="\t", header=0, index_col=0)
        if df.columns != ["count"]:
            raise ValueError(
                "Invalid column names for counts file [{}]" "".format(self.name)
            )
        if len(df) == 0:
            raise ValueError("Empty counts file [{}]".format(self.name))
        label = None
        for elem in ELEMENT_LABELS:
            if elem in self.labels:
                label = elem
                break
        if label is None:
            raise ValueError("No valid element labels [{}]".format(self.name))
        key = "/raw/{}/counts".format(label)
        self.store.put(
            key, df.astype(np.int32), format="table", data_columns=df.columns
        )

    def counts_from_file(self, fname):
        """Get raw counts from a counts file instead of FASTQ_ file.

        The ``'/raw/<element>/counts'`` table will be populated using the given
        input file. The input file should be a two-column file readable by
        ``pandas`` as a series or two-column dataframe or an Enrich2 HDF5 file.

        If the input file is a two-column file, the index will be checked using
        the SeqLib's ``validate_index()`` method.

        If the input file is an HDF5 file, the entire set of ``'/raw'`` tables
        will be copied over, with the metadata intact.
        """
        if not os.path.exists(fname):
            raise IOError("Counts file '{}' not found [{}]" "".format(fname, self.name))
        elif os.path.splitext(fname)[-1].lower() in (".h5"):
            self.counts_from_file_h5(self.counts_file)
        elif os.path.splitext(fname)[-1].lower() in (".txt", ".tsv", ".csv"):
            self.counts_from_file_tsv(self.counts_file)
        else:
            raise ValueError(
                "Unrecognized counts file extension for '{}' "
                "[{}]".format(fname, self.name)
            )




import logging
import sys
from .seqlib import SeqLib
from fqfa import open_compressed, parse_fastq_reads, has_fastq_ext


class BarcodeSeqLib(SeqLib):
    """
    Class for count data from barcoded sequencing libraries. Designed for
    barcode-only scoring or as a parent class for
    :py:class:`~seqlib.barcodevariant.BcvSeqLib` and
    :py:class:`~seqlib.barcodeid.BcidSeqLib`.
    """

    treeview_class_name = "Barcode SeqLib"

    def __init__(self):
        # Init step handled by VariantSeqLib's init for Barcode-variant
        if type(self).__name__ != "BcvSeqLib":
            SeqLib.__init__(self)
        self.reads = None
        self.reverse_complement_reads = False
        self.trim_start = 1
        self.trim_length = sys.maxsize
        self.barcode_min_count = 0
        self.add_label("barcodes")
        self.logger = logging.getLogger("{}.{}".format(__name__, self.__class__))

    def configure(self, cfg):
        """
        Set up the object using the config object *cfg*, usually derived from
        a ``.json`` file.
        """
        SeqLib.configure(self, cfg)
        self.logger = logging.getLogger(
            "{}.{} - {}".format(__name__, self.__class__.__name__, self.name)
        )

        # handle non-FASTQ config options
        try:
            if "min count" in cfg["barcodes"]:
                self.barcode_min_count = int(cfg["barcodes"]["min count"])
        except KeyError as key:
            raise KeyError("Missing required config value {}".format(key), self.name)

        # if counts are specified, copy them later
        # else handle the FASTQ config options and check the files
        if self.counts_file is None:
            self.configure_fastq(cfg)
            try:
                if not has_fastq_ext(self.reads):
                    raise ValueError(
                        "FASTQ file error: unrecognized file extension", self.name
                    )
            except IOError as fqerr:
                raise IOError("FASTQ file error: {}".format(fqerr), self.name)

    def serialize(self):
        """
        Format this object (and its children) as a config object suitable for
        dumping to a config file.
        """
        cfg = SeqLib.serialize(self)

        cfg["barcodes"] = dict()
        if self.barcode_min_count > 0:
            cfg["barcodes"]["min count"] = self.barcode_min_count

        cfg["fastq"] = self.serialize_fastq()

        return cfg

    def configure_fastq(self, cfg):
        """
        Set up the object's FASTQ_ file handling and filtering options.
        """
        try:
            self.reads = cfg["fastq"]["reads"]
            self.reverse_complement_reads = cfg["fastq"]["reverse"]

            if "start" in cfg["fastq"]:
                self.trim_start = cfg["fastq"]["start"]

            if "length" in cfg["fastq"]:
                self.trim_length = cfg["fastq"]["length"]

            self.filters = cfg["fastq"]["filters"]
        except KeyError as key:
            raise KeyError("Missing required config value {}".format(key), self.name)

    def serialize_fastq(self):
        """
        Serialize this object's FASTQ_ file handling and filtering options.
        """
        fastq = {
            "reads": self.reads,
            "reverse": self.reverse_complement_reads,
            "filters": self.serialize_filters(),
        }
        if self.trim_start > 1:
            fastq["start"] = self.trim_start

        if self.trim_length < sys.maxsize:
            fastq["length"] = self.trim_length

        return fastq

    def counts_from_reads(self):
        """
        Reads the forward or reverse FASTQ_ file (reverse reads are
        reverse-complemented), performs quality-based filtering, and counts
        the barcodes.

        Barcode counts after read-level filtering are stored under
        ``"/raw/barcodes/counts"``.
        """
        df_dict = dict()

        filter_flags = dict()
        for key in self.filters:
            filter_flags[key] = False

        # count all the barcodes
        self.logger.info("Counting barcodes")
        with open_compressed(self.reads) as handle:
            for fq in parse_fastq_reads(handle):
                fq.trim(start=self.trim_start, end=self.trim_start + self.trim_length -1)
                if self.reverse_complement_reads:
                    fq.reverse_complement()

                if self.read_quality_filter(fq):  # passed filtering
                    try:
                        df_dict[fq.sequence.upper()] += 1
                    except KeyError:
                        df_dict[fq.sequence.upper()] = 1

        self.save_counts("barcodes", df_dict, raw=True)
        del df_dict

    def calculate(self):
        """
        Counts the barcodes from the FASTQ file or from the provided counts
        file depending on the config.

        Barcodes that pass the minimum count
        filtering are stored under ``"/main/barcodes/counts"``.

        If ``"/main/barcodes/counts"`` already exists, those will be used
        instead of re-counting.
        """
        if self.check_store("/main/barcodes/counts"):
            return

        # no raw counts present
        if not self.check_store("/raw/barcodes/counts"):
            if self.counts_file is not None:
                self.counts_from_file(self.counts_file)
            else:
                self.counts_from_reads()

        if len(self.labels) == 1:  # only barcodes
            self.save_filtered_counts("barcodes", "count >= self.barcode_min_count")
            self.save_filter_stats()




from .barcode import BarcodeSeqLib
from .barcodevariant import BcvSeqLib
from .barcodeid import BcidSeqLib
from .basic import BasicSeqLib
from .idonly import IdOnlySeqLib
from .overlap import OverlapSeqLib
from .config_check import seqlib_type
from .storemanager import StoreManager
import os
import pandas as pd
import numpy as np
import logging
import statsmodels.api as sm
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import scipy.stats as stats
from .sfmap import sfmap_plot
from .plots import (
    fit_axes,
    fit_axes_text,
    volcano_plot,
    configure_axes,
    plot_colors,
    weights_plot,
)
from .constants import WILD_TYPE_VARIANT, SYNONYMOUS_VARIANT
from .variant import protein_variant
from .dataframe import singleton_dataframe


#: map class names to class definitions to avoid use of globals()
SEQLIB_CLASSES = {
    "BarcodeSeqLib": BarcodeSeqLib,
    "BcvSeqLib": BcvSeqLib,
    "BcidSeqLib": BcidSeqLib,
    "BasicSeqLib": BasicSeqLib,
    "IdOnlySeqLib": IdOnlySeqLib,
    "OverlapSeqLib": OverlapSeqLib,
}


def regression_apply(row, timepoints, weighted):
    """
    :py:meth:`pandas.DataFrame.apply` apply function for calculating
    enrichment using linear regression. If *weighted* is ``True`` perform
    weighted least squares; else perform ordinary least squares.

    Weights for weighted least squares are included in *row*.

    Returns a :py:class:`pandas.Series` containing regression coefficients,
    residuals, and statistics.
    """
    # retrieve log ratios from the row
    y = row[["L_{}".format(t) for t in timepoints]]

    # re-scale the x's to fall within [0, 1]
    xvalues = [x / float(max(timepoints)) for x in timepoints]

    # perform the fit
    X = sm.add_constant(xvalues)  # fit intercept
    if weighted:
        W = row[["W_{}".format(t) for t in timepoints]]
        fit = sm.WLS(y, X, weights=W).fit()
    else:
        fit = sm.OLS(y, X).fit()

    # re-format as a data frame row
    values = np.concatenate(
        [fit.params, [fit.bse["x1"], fit.tvalues["x1"], fit.pvalues["x1"]], fit.resid]
    )
    index = ["intercept", "slope", "SE_slope", "t", "pvalue_raw"] + [
        "e_{}".format(t) for t in timepoints
    ]
    return pd.Series(data=values, index=index)


class Selection(StoreManager):
    """
    Class for a single selection replicate, consisting of multiple
    timepoints. This class coordinates :py:class:`~seqlib.seqlib.SeqLib`
    objects.
    """

    store_suffix = "sel"
    treeview_class_name = "Selection"

    def __init__(self):
        StoreManager.__init__(self)
        self.libraries = dict()
        self.barcode_maps = dict()
        self._wt = None
        self.logger = logging.getLogger("{}.{}".format(__name__, self.__class__))

    def _children(self):
        """
        Return the :py:class:`~seqlib.seqlib.SeqLib` objects as a list,
        sorted by timepoint and then by name.
        """
        libs = list()
        for tp in self.timepoints:
            libs.extend(sorted(self.libraries[tp], key=lambda x: x.name))
        return libs

    def remove_child_id(self, tree_id):
        """
        Remove the reference to a :py:class:`~seqlib.seqlib.SeqLib` with
        Treeview id *tree_id*. Deletes empty time points.
        """
        empty = None
        for tp in self.libraries:
            tp_ids = [lib.treeview_id for lib in self.libraries[tp]]
            if tree_id in tp_ids:
                del self.libraries[tp][tp_ids.index(tree_id)]
                if len(self.libraries[tp]) == 0:
                    empty = tp
                break  # found the id, stop looking
        if empty is not None:
            del self.libraries[empty]

    @property
    def timepoints(self):
        return sorted(self.libraries.keys())

    @property
    def wt(self):
        if self.has_wt_sequence():
            if self._wt is None:
                self._wt = self.children[0].wt.duplicate(self.name)
            return self._wt
        else:
            if self._wt is not None:
                raise ValueError(
                    "Selection should not contain wild type sequence [{}]".format(
                        self.name
                    )
                )
            else:
                return None

    def configure(self, cfg, configure_children=True):
        """
        Set up the :py:class:`~selection.Selection` using the *cfg* object,
        usually from a ``.json`` configuration file.

        If *configure_children* is false, do not configure the children in
        *cfg*.
        """
        StoreManager.configure(self, cfg)
        self.logger = logging.getLogger(
            "{}.{} - {}".format(__name__, self.__class__.__name__, self.name)
        )
        if configure_children:
            if "libraries" not in cfg:
                raise KeyError(
                    "Missing required config value {} [{}]".format(
                        "libraries", self.name
                    )
                )

            for lib_cfg in cfg["libraries"]:
                libtype = seqlib_type(lib_cfg)
                if libtype is None:
                    raise ValueError("Unrecognized SeqLib config")
                elif libtype in ("BcvSeqLib", "BcidSeqLib"):
                    lib = SEQLIB_CLASSES[libtype]()
                    # don't re-parse the barcode maps if possible
                    mapfile = lib_cfg["barcodes"]["map file"]
                    if mapfile in list(self.barcode_maps.keys()):
                        lib.configure(lib_cfg, barcode_map=self.barcode_maps[mapfile])
                    else:
                        lib.configure(lib_cfg)
                        self.barcode_maps[mapfile] = lib.barcode_map
                    self.add_child(lib)
                else:
                    # requires that the SeqLib derived classes be imported into the
                    # module namespace using "from x import y" style
                    lib = SEQLIB_CLASSES[libtype]()
                    lib.configure(lib_cfg)
                    self.add_child(lib)

    def validate(self):
        """
        Raises an informative ``ValueError`` if the time points in the analysis are not suitable.

        Calls validate method on all child SeqLibs.
        """
        # check the time points
        if 0 not in self.timepoints:
            raise ValueError("Missing timepoint 0 [{}]".format(self.name))
        if self.timepoints[0] != 0:
            raise ValueError("Invalid negative timepoint [{}]".format(self.name))
        if len(self.timepoints) < 2:
            raise ValueError("Multiple timepoints required [{}]".format(self.name))
        elif len(self.timepoints) < 3 and self.scoring_method in ("WLS", "OLS"):
            raise ValueError(
                "Insufficient number of timepoints for regression scoring [{}]".format(
                    self.name
                )
            )

        # check the wild type sequences
        if self.has_wt_sequence():
            for child in self.children[1:]:
                if self.children[0].wt != child.wt:
                    self.logger.warning("Inconsistent wild type sequences")
                    break

        # check that we're not doing wild type normalization on something with no wild type
        # if not self.has_wt_sequence() and self.logr_method == "wt":
        #    raise ValueError("No wild type sequence for wild type normalization [{}]".format(self.name))

        # validate children
        for child in self.children:
            child.validate()

    def serialize(self):
        """
        Format this object (and its children) as a config object suitable for dumping to a config file.
        """
        cfg = StoreManager.serialize(self)
        cfg["libraries"] = [child.serialize() for child in self.children]
        return cfg

    def add_child(self, child):
        if child.name in self.child_names():
            raise ValueError(
                "Non-unique sequencing library name '{}' [{}]".format(
                    child.name, self.name
                )
            )

        child.parent = self

        # add it to the libraries dictionary
        try:
            self.libraries[child.timepoint].append(child)
        except KeyError:
            self.libraries[child.timepoint] = [child]

    def is_barcodevariant(self):
        """
        Return ``True`` if all :py:class:`~seqlib.seqlib.SeqLib` in the
        :py:class:`~selection.Selection` are
        :py:class:`~barcodevariant.BcvSeqLib` objects with
        the same barcode map, else ``False``.
        """
        return (
            all(isinstance(lib, BcvSeqLib) for lib in self.children)
            and len(list(self.barcode_maps.keys())) == 1
        )

    def is_barcodeid(self):
        """
        Return ``True`` if all :py:class:`~seqlib.SeqLib` in the
        :py:class:`~selection.Selection` are
        :py:class:`~barcodeid.BcidSeqLib` objects with
        the same barcode map, else ``False``.
        """
        return (
            all(isinstance(lib, BcidSeqLib) for lib in self.children)
            and len(list(self.barcode_maps.keys())) == 1
        )

    def is_coding(self):
        """
        Return ``True`` if the all :py:class:`~seqlib.seqlib.SeqLib` in the
        :py:class:`~selection.Selection` count protein-coding variants, else
        ``False``.
        """
        return all(x.is_coding() for x in self.children)

    def has_wt_sequence(self):
        """
        Return ``True`` if the all :py:class:`~seqlib.seqlib.SeqLib` in the
        :py:class:`~selection.Selection` have a wild type sequence, else
        ``False``.
        """
        return all(x.has_wt_sequence() for x in self.children)

    def merge_counts_unfiltered(self, label):
        """
        Counts :py:class:`~seqlib.seqlib.SeqLib` objects and tabulates counts
        for each timepoint. :py:class:`~seqlib.seqlib.SeqLib` objects from
        the same timepoint are combined by summing the counts.

        Stores the unfiltered counts under ``/main/label/counts_unfiltered``.
        """
        if self.check_store("/main/{}/counts_unfiltered".format(label)):
            return

        # calculate counts for each SeqLib
        self.logger.info("Counting for each time point ({})".format(label))
        for lib in self.children:
            lib.calculate()

        # combine all libraries for a given timepoint
        self.logger.info("Aggregating SeqLib data")

        destination = "/main/{}/counts_unfiltered".format(label)
        if destination in list(self.store.keys()):
            # need to remove the current destination table because we are using append
            # append is required because it takes the "min_itemsize" argument, and put doesn't
            self.logger.info("Replacing existing '{}'".format(destination))
            self.store.remove(destination)

        # seqlib count table name for this element type
        lib_table = "/main/{}/counts".format(label)

        # create an index of all elements in the analysis
        complete_index = pd.Index([])
        for tp in self.timepoints:
            for lib in self.libraries[tp]:
                complete_index = complete_index.union(
                    pd.Index(lib.store.select_column(lib_table, "index"))
                )
        self.logger.info(
            "Created shared index for count data ({} {})".format(
                len(complete_index), label
            )
        )

        # min_itemsize value
        max_index_length = complete_index.map(len).max()

        # perform operation in chunks
        tp_frame = None
        for i in range(0, len(complete_index), self.chunksize):
            # don't duplicate the index if the chunksize is large
            if self.chunksize < len(complete_index):
                index_chunk = complete_index[i : i + self.chunksize]
            else:
                index_chunk = complete_index
            self.logger.info(
                "Merging counts for chunk {} ({} rows)".format(
                    i / self.chunksize + 1, len(index_chunk)
                )
            )

            for tp in self.timepoints:
                c = self.libraries[tp][0].store.select(lib_table, "index = index_chunk")
                for lib in self.libraries[tp][1:]:
                    c = c.add(
                        lib.store.select(lib_table, "index = index_chunk"), fill_value=0
                    )
                c.columns = ["c_{}".format(tp)]
                if tp == 0:
                    tp_frame = c
                else:
                    tp_frame = tp_frame.join(c, how="outer")

            # save the unfiltered counts
            if "/main/{}/counts_unfiltered".format(label) not in self.store:
                self.store.append(
                    "/main/{}/counts_unfiltered".format(label),
                    tp_frame.astype(float),
                    min_itemsize={"index": max_index_length},
                    data_columns=list(tp_frame.columns),
                )
            else:
                self.store.append(
                    "/main/{}/counts_unfiltered".format(label), tp_frame.astype(float)
                )

    def filter_counts(self, label):
        """
        Converts unfiltered counts stored in ``/main/label/counts_unfiltered``
        into filtered counts calculated from complete cases (elements with a
        non-zero count in each time point).

        For the most basic element type (variant or barcode, depending on the
        experimental design), the result of this operation simply drops any
        rows that have missing counts. For other element types, such as
        synonymous variants, the counts are re-aggregated using only the
        complete cases in the underlying element type.
        """
        if (self.is_barcodeid() or self.is_barcodevariant()) and label != "barcodes":
            # calculate proper combined counts
            # df = self.store.select("/main/barcodes/counts") # this should exist because of the order of label calculations
            # redo the barcode->variant/id mapping with the filtered counts
            # NOT YET IMPLEMENTED
            df = self.store.select(
                "/main/{}/counts_unfiltered".format(label)
            )  # just do this for now
        else:
            df = self.store.select("/main/{}/counts_unfiltered".format(label))
        df.dropna(axis="index", how="any", inplace=True)
        self.store.put(
            "/main/{}/counts".format(label),
            df.astype(int),
            format="table",
            data_columns=df.columns,
        )

    def combine_barcode_maps(self):
        if self.check_store("/main/barcodemap"):
            return

        bcm = None
        for lib in self.children:
            if bcm is None:
                bcm = lib.store["/raw/barcodemap"]
            else:
                bcm = bcm.join(
                    lib.store["/raw/barcodemap"], rsuffix=".drop", how="outer"
                )
                new = bcm.loc[pd.isnull(bcm)["value"]]
                bcm.loc[new.index, "value"] = new["value.drop"]
                bcm.drop("value.drop", axis="columns", inplace=True)
        bcm.sort_values("value", inplace=True)
        self.store.put(
            "/main/barcodemap", bcm, format="table", data_columns=bcm.columns
        )

    def calculate(self):
        """
        Wrapper method to calculate counts and enrichment scores
        for all data in the :py:class:`~selection.Selection`.
        """
        if len(self.labels) == 0:
            raise ValueError(
                "No data present across all sequencing libraries [{}]".format(self.name)
            )
        for label in self.labels:
            self.merge_counts_unfiltered(label)
            self.filter_counts(label)
        if self.is_barcodevariant() or self.is_barcodeid():
            self.combine_barcode_maps()
        if self.scoring_method == "counts":
            pass
        elif self.scoring_method == "ratios":
            for label in self.labels:
                self.calc_ratios(label)
        elif self.scoring_method == "simple":
            for label in self.labels:
                self.calc_simple_ratios(label)
        elif self.scoring_method in ("WLS", "OLS"):
            if len(self.timepoints) <= 2:
                raise ValueError(
                    "Regression-based scoring requires three or more time points."
                )
            for label in self.labels:
                self.calc_log_ratios(label)
                if self.scoring_method == "WLS":
                    self.calc_weights(label)
                self.calc_regression(label)
        else:
            raise ValueError(
                'Invalid scoring method "{}" [{}]'.format(
                    self.scoring_method, self.name
                )
            )

        if self.scoring_method in ("ratios", "WLS", "OLS") and self.component_outliers:
            if self.is_barcodevariant() or self.is_barcodeid():
                self.calc_outliers("barcodes")
            if self.is_coding():
                self.calc_outliers("variants")

    def calc_simple_ratios(self, label):
        """
        Calculate simplified (original Enrich) ratios scores. This method does not produce standard errors.
        """
        if self.check_store("/main/{}/scores".format(label)):
            return

        self.logger.info("Calculating simple ratios ({})".format(label))
        c_last = "c_{}".format(self.timepoints[-1])
        df = self.store.select(
            "/main/{}/counts".format(label), "columns in ['c_0', c_last]"
        )

        # perform operations on the numpy values of the data frame for easier broadcasting
        ratios = (df[c_last].values.astype("float") / df[c_last].sum(axis="index")) / (
            df["c_0"].values.astype("float") / df["c_0"].sum(axis="index")
        )
        # make it a data frame again
        ratios = pd.DataFrame(data=ratios, index=df.index, columns=["ratio"])
        ratios["score"] = np.log2(ratios["ratio"])
        ratios["SE"] = np.nan
        ratios = ratios[["score", "SE", "ratio"]]  # re-order columns

        self.store.put(
            "/main/{}/scores".format(label),
            ratios,
            format="table",
            data_columns=ratios.columns,
        )

    def calc_ratios(self, label):
        """
        Calculate frequency ratios and standard errors between the last timepoint and the input.

        Ratios can be calculated using one of three methods:

        * wt
        * complete
        * full

        """
        if self.check_store("/main/{}/scores".format(label)):
            return

        self.logger.info("Calculating ratios ({})".format(label))
        c_last = "c_{}".format(self.timepoints[-1])
        df = self.store.select(
            "/main/{}/counts".format(label), "columns in ['c_0', c_last]"
        )

        if self.logr_method == "wt":
            if "variants" in self.labels:
                wt_label = "variants"
            elif "identifiers" in self.labels:
                wt_label = "identifiers"
            else:
                raise ValueError(
                    "Failed to use wild type log ratio method, suitable data table not present [{}]".format(
                        self.name
                    )
                )
            shared_counts = self.store.select(
                "/main/{}/counts".format(wt_label),
                "columns in ['c_0', c_last] & index='{}'".format(WILD_TYPE_VARIANT),
            )
            if len(shared_counts) == 0:  # wild type not found
                raise ValueError(
                    "Failed to use wild type log ratio method, wild type sequence not present [{}]".format(
                        self.name
                    )
                )
            shared_counts = shared_counts.values + 0.5
        elif self.logr_method == "complete":
            shared_counts = (
                self.store.select(
                    "/main/{}/counts".format(label), "columns in ['c_0', c_last]"
                )
                .sum(axis="index")
                .values
                + 0.5
            )
        elif self.logr_method == "full":
            shared_counts = (
                self.store.select(
                    "/main/{}/counts_unfiltered".format(label),
                    "columns in ['c_0', c_last]",
                )
                .sum(axis="index", skipna=True)
                .values
                + 0.5
            )
        else:
            raise ValueError(
                'Invalid log ratio method "{}" [{}]'.format(self.logr_method, self.name)
            )

        ratios = np.log(df[["c_0", c_last]].values + 0.5) - np.log(shared_counts)
        ratios = ratios[:, 1] - ratios[:, 0]  # selected - input

        ratios = pd.DataFrame(data=ratios, index=df.index, columns=["logratio"])

        shared_variance = np.sum(1.0 / shared_counts)
        ratios["variance"] = (
            np.sum(1.0 / (df[["c_0", c_last]].values + 0.5), axis=1) + shared_variance
        )

        ratios["score"] = ratios["logratio"]
        ratios["SE"] = np.sqrt(ratios["variance"])
        ratios = ratios[["score", "SE", "logratio", "variance"]]  # re-order columns

        self.store.put(
            "/main/{}/scores".format(label),
            ratios,
            format="table",
            data_columns=ratios.columns,
        )

    def calc_log_ratios(self, label):
        """
        Calculate the log ratios that will be fit using the linear model.
        """
        if self.check_store("/main/{}/log_ratios".format(label)):
            return

        self.logger.info("Calculating log ratios ({})".format(label))
        ratios = self.store.select("/main/{}/counts".format(label))
        index = ratios.index
        c_n = ["c_{}".format(x) for x in self.timepoints]
        ratios = np.log(ratios + 0.5)

        # perform operations on the numpy values of the data frame for easier broadcasting
        ratios = ratios[c_n].values
        if self.logr_method == "wt":
            if "variants" in self.labels:
                wt_label = "variants"
            elif "identifiers" in self.labels:
                wt_label = "identifiers"
            else:
                raise ValueError(
                    "Failed to use wild type log ratio method, suitable data table not present [{}]".format(
                        self.name
                    )
                )
            wt_counts = self.store.select(
                "/main/{}/counts".format(wt_label),
                "columns=c_n & index='{}'".format(WILD_TYPE_VARIANT),
            )
            if len(wt_counts) == 0:  # wild type not found
                raise ValueError(
                    "Failed to use wild type log ratio method, wild type sequence not present [{}]".format(
                        self.name
                    )
                )
            ratios = ratios - np.log(wt_counts.values + 0.5)
        elif self.logr_method == "complete":
            ratios = ratios - np.log(
                self.store.select("/main/{}/counts".format(label), "columns=c_n")
                .sum(axis="index")
                .values
                + 0.5
            )
        elif self.logr_method == "full":
            ratios = ratios - np.log(
                self.store.select(
                    "/main/{}/counts_unfiltered".format(label), "columns=c_n"
                )
                .sum(axis="index", skipna=True)
                .values
                + 0.5
            )
        else:
            raise ValueError(
                'Invalid log ratio method "{}" [{}]'.format(self.logr_method, self.name)
            )

        # make it a data frame again
        ratios = pd.DataFrame(
            data=ratios,
            index=index,
            columns=["L_{}".format(x) for x in self.timepoints],
        )
        self.store.put(
            "/main/{}/log_ratios".format(label),
            ratios,
            format="table",
            data_columns=ratios.columns,
        )

    def calc_weights(self, label):
        """
        Calculate the regression weights (1 / variance).
        """
        if self.check_store("/main/{}/weights".format(label)):
            return

        self.logger.info("Calculating regression weights ({})".format(label))
        variances = self.store.select("/main/{}/counts".format(label))
        c_n = ["c_{}".format(x) for x in self.timepoints]
        index = variances.index
        # perform operations on the numpy values of the data frame for easier broadcasting
        # variances = 1.0 / (variances[c_n].values + 0.5) + 1.0 / (variances[['c_0']].values + 0.5)
        variances = 1.0 / (variances[c_n].values + 0.5)
        if self.logr_method == "wt":
            if "variants" in self.labels:
                wt_label = "variants"
            elif "identifiers" in self.labels:
                wt_label = "identifiers"
            else:
                raise ValueError(
                    "Failed to use wild type log ratio method, suitable data table not present [{}]".format(
                        self.name
                    )
                )
            wt_counts = self.store.select(
                "/main/{}/counts".format(wt_label),
                "columns=c_n & index='{}'".format(WILD_TYPE_VARIANT),
            )
            if len(wt_counts) == 0:  # wild type not found
                raise ValueError(
                    "Failed to use wild type log ratio method, wild type sequence not present [{}]".format(
                        self.name
                    )
                )
            variances = variances + 1.0 / (wt_counts.values + 0.5)
        elif self.logr_method == "complete":
            variances = variances + 1.0 / (
                self.store.select("/main/{}/counts".format(label), "columns=c_n")
                .sum(axis="index")
                .values
                + 0.5
            )
        elif self.logr_method == "full":
            variances = variances + 1.0 / (
                self.store.select(
                    "/main/{}/counts_unfiltered".format(label), "columns=c_n"
                )
                .sum(axis="index", skipna=True)
                .values
                + 0.5
            )
        else:
            raise ValueError(
                'Invalid log ratio method "{}" [{}]'.format(self.logr_method, self.name)
            )
        variances = 1.0 / variances  # weights are reciprocal of variances
        # make it a data frame again
        variances = pd.DataFrame(
            data=variances,
            index=index,
            columns=["W_{}".format(x) for x in self.timepoints],
        )
        self.store.put(
            "/main/{}/weights".format(label),
            variances,
            format="table",
            data_columns=variances.columns,
        )

    def calc_regression(self, label):
        """
        Calculate least squares regression for *label*. If *weighted* is ``True``, calculates weighted least squares; else ordinary least squares.

        Regression results are stored in ``'/main/label/scores'``

        """
        if self.check_store("/main/{}/scores".format(label)):
            return
        elif "/main/{}/scores".format(label) in list(self.store.keys()):
            # need to remove the current keys because we are using append
            self.store.remove("/main/{}/scores".format(label))

        self.logger.info(
            "Calculating {} regression coefficients ({})".format(
                self.scoring_method, label
            )
        )
        # append is required because it takes the "min_itemsize" argument, and put doesn't
        longest = (
            self.store.select("/main/{}/log_ratios".format(label), "columns='index'")
            .index.map(len)
            .max()
        )
        chunk = 1
        if self.scoring_method == "WLS":
            for data in self.store.select_as_multiple(
                ["/main/{}/log_ratios".format(label), "/main/{}/weights".format(label)],
                chunksize=self.chunksize,
            ):
                self.logger.info(
                    "Calculating weighted least squares for chunk {} ({} rows)".format(
                        chunk, len(data.index)
                    )
                )
                result = data.apply(
                    regression_apply, args=[self.timepoints, True], axis="columns"
                )
                self.store.append(
                    "/main/{}/scores".format(label),
                    result,
                    min_itemsize={"index": longest},
                )
                chunk += 1
        elif self.scoring_method == "OLS":
            for data in self.store.select(
                "/main/{}/log_ratios".format(label), chunksize=self.chunksize
            ):
                self.logger.info(
                    "Calculating ordinary least squares for chunk {} ({} rows)".format(
                        chunk, len(data.index)
                    )
                )
                result = data.apply(
                    regression_apply, args=[self.timepoints, False], axis="columns"
                )
                self.store.append(
                    "/main/{}/scores".format(label),
                    result,
                    min_itemsize={"index": longest},
                )
                chunk += 1
        else:
            raise ValueError(
                'Invalid regression scoring method "{}" [{}]'.format(
                    self.scoring_method, self.name
                )
            )

        # need to read from the file, calculate percentiles, and rewrite it
        self.logger.info(
            "Calculating slope standard error percentiles ({})".format(label)
        )
        data = self.store["/main/{}/scores".format(label)]
        data["score"] = data["slope"]
        data["SE"] = data["SE_slope"]
        data["SE_pctile"] = [
            stats.percentileofscore(data["SE"], x, "weak") for x in data["SE"]
        ]
        data = data[
            [
                "score",
                "SE",
                "SE_pctile",
                "slope",
                "intercept",
                "SE_slope",
                "t",
                "pvalue_raw",
            ]
        ]  # reorder columns
        self.store.put(
            "/main/{}/scores".format(label),
            data,
            format="table",
            data_columns=data.columns,
        )

    def wt_plot(self, pdf):
        """
        Create a plot of the linear fit of the wild type variant.

        *pdf* is an open PdfPages instance.

        Only created for selections that use WLS or OLS scoring and have a wild type specified.
        Uses :py:func:`~plots.fit_axes` for the plotting.
        """
        self.logger.info("Creating wild type fit plot")

        # get the data and calculate log ratios
        if "variants" in self.labels:
            wt_label = "variants"
        elif "identifiers" in self.labels:
            wt_label = "identifiers"
        else:
            raise ValueError(
                "Invalid selection type for wild type fit plot [{}]".format(self.name)
            )
        data = self.store.select(
            "/main/{}/counts".format(wt_label),
            where='index = "{}"'.format(WILD_TYPE_VARIANT),
        ).iloc[0]
        sums = self.store["/main/{}/counts".format(wt_label)].sum(
            axis="index"
        )  # sum of complete cases (N')
        yvalues = np.log(data + 0.5) - np.log(sums + 0.5)
        xvalues = [tp / float(max(self.timepoints)) for tp in self.timepoints]

        # fit the line
        X = sm.add_constant(xvalues)  # fit intercept
        if self.scoring_method == "WLS":
            W = 1 / (1 / (data + 0.5) + 1 / (sums + 0.5))
            fit = sm.WLS(yvalues, X, weights=W).fit()
        elif self.scoring_method == "OLS":
            fit = sm.OLS(yvalues, X).fit()
        else:
            raise ValueError(
                'Invalid regression scoring method "{}" [{}]'.format(
                    self.scoring_method, self.name
                )
            )
        intercept, slope = fit.params
        slope_se = fit.bse["x1"]

        # make the plot
        fig, ax = plt.subplots()
        fig.set_tight_layout(True)
        fit_axes(ax, xvalues, yvalues, slope, intercept, xlabels=self.timepoints)
        fit_axes_text(ax, cornertext="Slope {:3.2f}\nSE {:.1f}".format(slope, slope_se))
        ax.set_title("Wild Type Shape\n{}".format(self.name))
        ax.set_ylabel("Log Ratio (Complete Cases)")

        pdf.savefig(fig)
        plt.close(fig)

    def se_pctile_plot(self, label, pdf):
        """
        Create plots of the linear fit of 21 selected variants, evenly distributed based on their standard error percentiles (0, 5, 10, ..., 100).

        *label* is the data label (barcode, variant, etc.)

        *pdf* is an open PdfPages instance.

        Uses :py:func:`~plots.fit_axes` for the plotting.
        """
        self.logger.info("Creating representative fit plots ({})".format(label))

        se_data = self.store.select(
            "/main/{}/scores".format(label),
            where="columns in ['slope', 'intercept', 'SE_pctile'] & index!='{}' & index!='{}'".format(
                WILD_TYPE_VARIANT, SYNONYMOUS_VARIANT
            ),
        )
        se_data.sort_values("SE_pctile", inplace=True)
        se_data.dropna(axis="index", how="any", inplace=True)

        indices = np.linspace(0, len(se_data.index) - 1, 21).astype("int")
        se_data = se_data.iloc[indices]

        # retrieves the whole DF because one case was hanging when trying to use select
        # totally unexplained, should fix later
        ratio_data = self.store.select("/main/{}/log_ratios".format(label)).loc[
            se_data.index
        ]

        fig, axarr = plt.subplots(7, 3, sharex="all", sharey="all")
        fig.subplots_adjust(
            hspace=0, wspace=0
        )  # eliminate white space between the subplots
        fig.set_size_inches((10, 17))
        fig.suptitle(
            "Representative {} Fits\n{} ({})".format(
                self.scoring_method, self.name, label.title()
            )
        )
        fig.subplots_adjust(top=0.958)  # eliminate white space after the title

        # turn off all tick labels
        plt.setp([ax.get_xticklabels() for ax in axarr.reshape(-1)], visible=False)
        plt.setp([ax.get_yticklabels() for ax in axarr.reshape(-1)], visible=False)

        # tick labels back on for some plots
        plt.setp(
            [ax.get_xticklabels() for ax in (axarr[-1, 0], axarr[-1, 2])], visible=True
        )
        plt.setp(
            [
                ax.get_yticklabels()
                for ax in (axarr[0, 0], axarr[2, 0], axarr[4, 0], axarr[6, 0])
            ],
            visible=True,
        )
        plt.setp(
            [ax.get_yticklabels() for ax in (axarr[1, -1], axarr[3, -1], axarr[5, -1])],
            visible=True,
        )
        plt.setp(
            [ax.yaxis for ax in (axarr[1, -1], axarr[3, -1], axarr[5, -1])],
            ticks_position="right",
        )

        # create the fit plots and add text to the individual plots
        for i, ax in enumerate(axarr.reshape(-1)):
            index = se_data.index[i]
            fit_axes(
                ax,
                xvalues=[x / float(max(self.timepoints)) for x in self.timepoints],
                yvalues=ratio_data.loc[index],
                slope=se_data.loc[index, "slope"],
                intercept=se_data.loc[index, "intercept"],
                xlabels=self.timepoints,
            )
            fit_axes_text(
                ax,
                cornertext="Slope {:3.2f}\nSE Pctile {:.1f}".format(
                    se_data.loc[index, "slope"], se_data.loc[index, "SE_pctile"]
                ),
                centertext=index,
            )

        # turn off the subplot axis labels
        [ax.set_xlabel("") for ax in axarr.reshape(-1)]
        [ax.set_ylabel("") for ax in axarr.reshape(-1)]

        # add x and y labels in the middle
        axarr[-1, 1].set_xlabel("Time Point")
        axarr[3, 0].set_ylabel("Log Ratio")

        pdf.savefig(fig, bbox_inches="tight")
        plt.close(fig)

    def timepoint_counts_plot(self, label, pdf):
        """
        Create barplot of the number of items counted for each time point.

        *label* is the data label (barcode, variant, etc.)

        *pdf* is an open PdfPages instance.
        """
        self.logger.info("Creating time point count plots ({})".format(label))

        counts = self.store["/main/{}/counts".format(label)].sum(axis="index")

        fig, ax = plt.subplots()
        fig.set_tight_layout(True)
        configure_axes(ax)

        xpos = np.arange(len(self.timepoints))
        width = 0.8
        ax.bar(xpos, counts, width, color=plot_colors["bright1"])
        ax.set_title("Total {}\n{}".format(label.title(), self.name))
        ax.set_ylabel("Count")
        ax.set_xlabel("Timepoint")
        ax.set_xticks(xpos + width / 2.0)
        ax.set_xticklabels(self.timepoints)

        pdf.savefig(fig)
        plt.close(fig)

    def volcano_plot(self, label, pdf, colors="YlGnBu_r", log_bins=True):
        """
        Create a volcano plot (p-value vs. functional score).

        *label* is the data label (barcode, variant, etc.)

        *pdf* is an open PdfPages instance.

        The p-values used are the regression p-values (p-value of non-zero slope). Due to the large number of points, we use a hexbin plot showing the density instead of a scatter plot.
        """
        self.logger.info("Creating volcano plot ({})".format(label))

        # get the data
        data = self.store.select(
            "/main/{}/scores".format(label), "columns=['score', 'pvalue_raw']"
        )
        volcano_plot(
            data,
            pdf,
            title="{} ({})".format(self.name, label.title()),
            colors=colors,
            log_bins=log_bins,
        )

    def make_plots(self):
        """
        Create plots for this entity.

        This function handles opening and closing the various PDF files for multi-page plots, as well as plotting similar plots for different data labels.
        """
        if self.plots_requested:
            self.logger.info("Creating plots")

            # counts per time point
            pdf = PdfPages(os.path.join(self.plot_dir, "timepoint_counts.pdf"))
            for label in self.labels:
                self.timepoint_counts_plot(label, pdf)
            pdf.close()

            # wild type shape
            if self.logr_method == "wt" and self.scoring_method in ("WLS", "OLS"):
                pdf = PdfPages(os.path.join(self.plot_dir, "wt_shape.pdf"))
                self.wt_plot(pdf)
                pdf.close()

            # regression weights
            if self.scoring_method == "WLS":
                pdf = PdfPages(os.path.join(self.plot_dir, "regression_weights.pdf"))
                for label in self.labels:
                    weights_plot(self, label, pdf)
                pdf.close()

            # linear fits by standard error percentile
            if self.scoring_method in ("WLS", "OLS"):
                pdf = PdfPages(os.path.join(self.plot_dir, "se_pctile.pdf"))
                for label in self.labels:
                    self.se_pctile_plot(label, pdf)
                pdf.close()

            # volcano plots
            # if self.scoring_method in ("WLS", "OLS", "ratios"):
            if self.scoring_method in ("WLS", "OLS") and "variants" in self.labels:
                pdf = PdfPages(os.path.join(self.plot_dir, "volcano.pdf"))
                for label in self.labels:
                    self.volcano_plot(label, pdf, log_bins=True)
                pdf.close()

            # library diversity for each time point (amino acid)
            if "synonymous" in self.labels:
                pdf = PdfPages(os.path.join(self.plot_dir, "diversity_aa.pdf"))
                for tp in self.timepoints:
                    self.sfmap_wrapper(
                        cname="c_{}".format(tp), pdf=pdf, coding=True, log10=True
                    )
                pdf.close()

            # library diversity for each time point (nucleotide)
            if "variants" in self.labels:
                pdf = PdfPages(os.path.join(self.plot_dir, "diversity_nt.pdf"))
                for tp in self.timepoints:
                    self.sfmap_wrapper(
                        cname="c_{}".format(tp), pdf=pdf, coding=False, log10=True
                    )
                pdf.close()

            # sequence-function maps
            if self.scoring_method != "counts":
                if "synonymous" in self.labels:
                    pdf = PdfPages(
                        os.path.join(self.plot_dir, "sequence_function_map_aa.pdf")
                    )
                    self.sfmap_wrapper(cname="score", pdf=pdf, coding=True)
                    pdf.close()
                if "variants" in self.labels:
                    pdf = PdfPages(
                        os.path.join(self.plot_dir, "sequence_function_map_nt.pdf")
                    )
                    self.sfmap_wrapper(cname="score", pdf=pdf, coding=False)
                    pdf.close()

        # SeqLib plots
        for lib in self.children:
            lib.make_plots()

    def write_tsv(self):
        """
        Write each table from the store to its own tab-separated file.

        Files are written to a ``tsv`` directory in the default output location.
        File names are the HDF5 key with ``'_'`` substituted for ``'/'``.
        """
        if self.tsv_requested:
            self.logger.info("Generating tab-separated output files")
            for k in list(self.store.keys()):
                self.write_table_tsv(k)
        for lib in self.children:
            lib.write_tsv()

    def synonymous_variants(self):
        """
        Populate and return a dictionary mapping synonymous variants to the
        list of associated variants in ``/main/variants/counts``.
        """
        mapping = dict()
        try:
            variants = self.store.select_column("/main/variants/counts", "index")
        except KeyError:
            raise KeyError("No variant counts found [{}]".format(self.name))
        for v in variants:
            pv = protein_variant(v)
            try:
                mapping[pv].append(v)
            except KeyError:
                mapping[pv] = [v]
        return mapping

    def sfmap_wrapper(self, cname, pdf, coding, log10=False):
        """
        Create a sequence function map for either scores or library diversity.

        Uses :py:func:`~sfmap.sfmap_plot` for the plotting.
        """
        plot_options = self.get_root().plot_options

        if cname.startswith("c_"):
            counts = True
        elif cname == "score":
            counts = False
        else:
            raise ValueError("Invalid sequence-function map data column.")

        if coding:
            label = "amino acid"
        else:
            label = "nucleotide"

        if counts:
            self.logger.info("Creating diversity map ({})".format(label))
        else:
            self.logger.info("Creating sequence-function map ({})".format(label))

        # build the data frame name and get the data
        df_name = "/main/"
        if coding:
            df_name += "synonymous/"
        else:
            df_name += "variants/"
        if counts:
            df_name += "counts_unfiltered"
        else:
            df_name += "scores"
        if plot_options is not None:
            data, wtseq = singleton_dataframe(
                self.store[df_name][cname],
                self.wt,
                coding=coding,
                aa_list=plot_options["aa_list"],
            )
        else:
            data, wtseq = singleton_dataframe(
                self.store[df_name][cname], self.wt, coding=coding
            )
        if counts:
            data_se = None
        else:
            if plot_options is not None:
                data_se, _ = singleton_dataframe(
                    self.store[df_name]["SE"],
                    self.wt,
                    coding=coding,
                    aa_list=plot_options["aa_list"],
                )
            else:
                data_se, _ = singleton_dataframe(
                    self.store[df_name]["SE"], self.wt, coding=coding
                )

        # format the title
        if coding:
            title = "Amino Acid"
        else:
            title = "Nucleotide"
        if counts:
            title += " Diversity Map\n{} (Time {})".format(
                self.name, cname[2:]
            )  # trim the "c_"
        else:
            if self.scoring_method in ("WLS", "OLS"):
                title += " Sequence-Function Map\n{} ({} Slope)".format(
                    self.name, self.scoring_method
                )
            elif self.scoring_method == "ratios":
                title += " Sequence-Function Map\n{} ({})".format(
                    self.name, "Enrich2 Ratio"
                )
            elif self.scoring_method == "simple":
                title += " Sequence-Function Map\n{} ({})".format(
                    self.name, "Simplified Ratio"
                )
            else:
                raise ValueError("Invalid scoring method", self.name)

        if counts and log10:
            style = "logcounts"
        elif counts:
            style = "counts"
        else:
            style = "scores"

        if plot_options is not None:
            sfmap_plot(
                df=data,
                pdf=pdf,
                style=style,
                df_se=data_se,
                dimensions="tall",
                wt=wtseq,
                title=title,
                aa_list=plot_options["aa_list"],
                aa_label_groups=plot_options["aa_label_groups"],
            )
        else:
            sfmap_plot(
                df=data,
                pdf=pdf,
                style=style,
                df_se=data_se,
                dimensions="tall",
                wt=wtseq,
                title=title,
            )

    def barcodemap_mapping(self):
        mapping = dict()
        for bc, v in self.store["/main/barcodemap"].iterrows():
            v = v["value"]
            try:
                mapping[v].update([bc])
            except KeyError:
                mapping[v] = {bc}
        return mapping

    def calc_outliers(self, label, minimum_components=4, log_chunksize=20000):
        """
        Test whether an element's individual components have significantly different
        scores from the element. Results are stored in ``'/main/<label>/outliers'``.

        Args:
            label (str): label for the component (``'variants'`` or ``'barcodes'``)

            minimum_components (int): minimum number of componenents required for any statistics to be calculated

            log_chunksize (int): will output a log message every *n* rows

        """
        if self.check_store("/main/{}/outliers".format(label)):
            return

        if label == "variants":
            label2 = "synonymous"
        elif label == "barcodes":
            if self.is_barcodevariant():
                label2 = "variants"
            elif self.is_barcodeid():
                label2 = "identifiers"
            else:
                # this should never happen
                raise AttributeError(
                    "Failed to identify parent label when calculating barcode outliers [{}]".format(
                        self.name
                    )
                )
        else:
            raise KeyError(
                "Invalid label '{}' for calc_outliers [{}]".format(label, self.name)
            )

        self.logger.info("Identifying outliers ({}-{})".format(label, label2))

        self.logger.info("Mapping {} to {}".format(label, label2))
        if label == "variants":
            mapping = self.synonymous_variants()
        elif label == "barcodes":
            mapping = self.barcodemap_mapping()
        else:
            raise KeyError(
                "Invalid label '{}' for calc_outliers [{}]".format(label, self.name)
            )

        # get the scores
        df1 = self.store.select(
            "/main/{}/scores".format(label), "columns=['score', 'SE']"
        )
        df2 = self.store.select(
            "/main/{}/scores".format(label2), "columns=['score', 'SE']"
        )

        # pre-calculate variances
        df1["var"] = df1["SE"] ** 2
        df1.drop("SE", axis=1, inplace=True)
        df2["var"] = df2["SE"] ** 2
        df2.drop("SE", axis=1, inplace=True)

        # set up empty results DF
        result_df = pd.DataFrame(
            np.nan, index=df1.index, columns=["z", "pvalue_raw", "parent"]
        )

        # because this step can be slow, output chunk-style logging messages
        # pre-calculate the lengths for the log messages
        log_chunksize_list = [log_chunksize] * (len(df2) / log_chunksize) + [
            len(df2) % log_chunksize
        ]
        log_chunk = 1

        for i, x in enumerate(df2.index):
            if i % log_chunksize == 0:
                self.logger.info(
                    "Calculating outlier p-values for chunk {} ({} rows) ({}-{})".format(
                        log_chunk, log_chunksize_list[log_chunk - 1], label, label2
                    )
                )
                log_chunk += 1
            try:
                components = df1.loc[mapping[x]].dropna(axis="index", how="all")
            except KeyError:
                # none of the components were in the index
                continue
            if len(components.index) >= minimum_components:
                for c in components.index:
                    zvalue = np.absolute(
                        df2.loc[x, "score"] - df1.loc[c, "score"]
                    ) / np.sqrt(df2.loc[x, "var"] + df1.loc[c, "var"])
                    result_df.loc[c, "z"] = zvalue
                    result_df.loc[c, "pvalue_raw"] = 2 * stats.norm.sf(zvalue)
                    result_df.loc[c, "parent"] = x
        if WILD_TYPE_VARIANT in result_df.index:
            result_df.loc[WILD_TYPE_VARIANT, "z"] = np.nan
            result_df.loc[WILD_TYPE_VARIANT, "pvalue_raw"] = np.nan
        result_df["z"] = result_df["z"].astype(float)
        result_df["pvalue_raw"] = result_df["pvalue_raw"].astype(float)

        self.store.put(
            "/main/{}/outliers".format(label),
            result_df,
            format="table",
            data_columns=result_df.columns,
        )




import collections
import logging
import numpy as np
import matplotlib.pyplot as plt
import scipy.stats
from statsmodels.nonparametric.kde import KDEUnivariate

logger = logging.getLogger(__name__)

tick_params = {"labelsize": 6, "pad": 2}

label_params = {"fontsize": 8, "labelpad": 2}

legend_params = {"fontsize": 8}

letter_params = {
    "fontsize": 10,
    "weight": "bold",
    "horizontalalignment": "left",
    "verticalalignment": "center",
}

#: Default colors for Enrich2 plots.
#: Standard defaults are based on ColorBrewer Set1 and Pastel1.
plot_colors = {
    "bright1": "#377eb8",
    "bright2": "#4daf4a",
    "bright3": "#ff7f00",
    "bright4": "#984ea3",
    "bright5": "#e41a1c",
    "bright6": "#a65628",
    "pastel1": "#b3cde3",
    "pastel2": "#ccebc5",
    "pastel3": "#fed9a6",
    "pastel4": "#decbe4",
    "pastel5": "#fbb4ae",
    "pastel6": "#e5d8bd",
    "accent1": "#000000",
    "accent2": "#545454",
    "accent3": "#bebebe",
    "accent4": "#999999",
    "missing": "#808080",
}


#: Default matplotlib color maps.
#: ``'diverging'`` is used for sequence-function maps.
#: ``'sequential'`` is used for library diversity plots.
#: ``'hexbin'`` is used for volcano plots.
plot_cmaps = {"diverging": "RdBu_r", "sequential": "BuPu", "hexbin": "YlGnBu_r"}


#: Default number of histogram bins.
DEFAULT_BINS = 30

#: Default number of hexbins in each direction (x and y).
DEFAULT_GRIDSIZE = 50


def configure_axes(ax, xgrid=False):
    """
    Apply the standard Enrich2 style to the axes *ax*. This includes invisible 
    tick marks, horizontal (and optionally vertical) grey grid lines at major 
    tick intervals, and scientific notation on numerical axes.

    *ax* is the axes object that will have its properties set.

    *xgrid* sets the drawing of vertical grid lines that match the horizontal 
    grid lines.

    """
    # turn off all tick marks
    ax.tick_params(
        axis="both", which="both", bottom=False, top=False, left=False, right=False
    )
    # turn on y-axis grid (horizontal lines)
    ax.yaxis.grid(color=plot_colors["accent3"], linestyle="-")
    # optionally turn on x-axis grid (vertical lines)
    if xgrid:  # for scatterplots, hexbin
        ax.xaxis.grid(color=plot_colors["accent3"], linestyle="-")
    else:  # for barplots, histograms, etc.
        ax.xaxis.grid(False)
    # put the axis and grid behind any plot elements
    ax.set_axisbelow(True)
    # use scientific notation for large or small tick label values
    ax.ticklabel_format(style="sci", scilimits=(-3, 3))


def fit_axes(ax, xvalues, yvalues, slope, intercept, xlabels=None):
    """
    Create a scatterplot with fit line for a particular element's 
    regression-based score. This is used for both WLS and OLS scoring methods.

    *ax* is the axes object used for plotting.

    *xvalues* are the regression's x-values (normalized time points).

    *yvalues* are the regression's y-values (log ratios).

    *slope* is the slope of the fit line.

    *intercept* is the intercept of the fit line.

    *xlabels* are optional tick labels for the x-axis (usually the 
    pre-normalization time points). If *xlabels* is ``None``, the *xvalues* 
    will be used as tick labels.

    """
    # set appropriate x axis limits
    xmin = sorted(xvalues)[0]
    xmax = sorted(xvalues)[-1]
    xlim_margin = (xmax - xmin) * 0.1
    ax.set_xlim([xmin - xlim_margin, xmax + xlim_margin])

    # add axis labels
    ax.set_xlabel("Time Point")
    ax.set_ylabel("Log Ratio")

    # add x tick labels
    ax.set_xticks(xvalues)
    if xlabels is None:
        ax.set_xticklabels(xvalues)
    else:
        ax.set_xticklabels(xlabels)

    # plot the log ratios
    ax.scatter(xvalues, yvalues, color=plot_colors["bright1"], marker="D")

    # plot residual lines
    fit_yvalues = [x * slope + intercept for x in xvalues]
    for x, y1, y2 in zip(xvalues, yvalues, fit_yvalues):
        ax.plot(
            (x, x), (y1, y2), color=plot_colors["bright1"], linestyle="-", alpha=0.5
        )

    # plot the fit line
    ax.plot(
        ax.get_xlim(),
        [x * slope + intercept for x in ax.get_xlim()],
        color=plot_colors["bright3"],
        linestyle="--",
    )


def fit_axes_text(ax, cornertext=None, centertext=None, centerwrap=40):
    """
    Add text to a fit plot in either the bottom left corner and/or the top
    middle.

    *ax* is the axes object containing the fit plot.

    *cornertext* is displayed in the lower left corner of the plot. Typically
    this includes details about the fit (slope, standard error, etc.). If it
    is ``None``, no text is displayed.

    *centertext* is displayed in a smaller font in the upper center of the
    plot. Typically this is the HGVS_ variant string or the barcode for the
    element whose fit is plotted. The string is split on commas into a
    multi-line string as required given its length (see *centerwrap*).

    *centerwrap* is the maximum length of each line in the *centertext* string.

    """
    if cornertext is not None:
        ax.text(
            0.05,
            0.05,
            cornertext,
            horizontalalignment="left",
            verticalalignment="bottom",
            transform=ax.transAxes,
        )

    if centertext is not None:
        splits = centertext.split(",")
        splits = [x.strip() for x in splits]
        length = len(splits[0])
        text = splits[0]
        for x in splits[1:]:
            if length + len(x) < centerwrap:
                text += ", " + x
                length += 2 + len(x)
            else:
                text += ",\n" + x
                length = len(x)
        ax.text(
            0.5,
            0.95,
            text,
            horizontalalignment="center",
            verticalalignment="top",
            transform=ax.transAxes,
            fontsize=8,
        )


def volcano_plot(df, pdf, title=None, colors=None, log_bins=True, logp_max=10):
    """
    Create a volcano plot of p-values vs. functional scores.

    *df* is a DataFrame containing ``'score'`` and ``'pvalue_raw'`` columns.
        Each element (barcode, variant, etc.) has its own row. Other columns
        will be ignored.

    *pdf* is the open PdfPages file object.

    *title* is the plot title. If ``None``, no title will be displayed.

    *colors* is the name of the `matplotlib cmap`_ used.
        If *colors* is ``None``, the default hexbin cmap will be used.

    *log_bins* sets whether or not the counts in each hexbin should be log10
    transformed.

    *logp_max* sets the maximum minus log-transformed p-value that will be
        plotted. p-values smaller than this will be set to the given max.

    """
    df = df.loc[:, ("pvalue_raw", "score")]
    df.dropna(axis="index", how="any", inplace=True)
    df["pvalue_raw"] = -np.log10(df["pvalue_raw"])
    df.loc[df["pvalue_raw"] > logp_max, "pvalue_raw"] = logp_max  # set max

    # create the figure
    fig, ax = plt.subplots()
    fig.set_tight_layout(True)
    configure_axes(ax, xgrid=True)

    # make the plot
    if log_bins:
        bins = "log"
    else:
        bins = None
    hexbin = ax.hexbin(
        x=df["score"].values,
        y=df["pvalue_raw"].values,
        cmap=colors,
        bins=bins,
        mincnt=1,
        edgecolor="none",
        gridsize=DEFAULT_GRIDSIZE,
    )

    # set the labels
    if title is not None:
        ax.set_title(title)
    ax.set_ylabel("-log10(raw p-value)")
    ax.set_xlabel("Score")

    # space the boundaries out so the points aren't on the edges
    ax.set_xlim([1.05 * df["score"].min(), 1.05 * df["score"].max()])
    ax.set_ylim([0.0, 1.05 * df["pvalue_raw"].max()])

    # create color bar
    cbar = fig.colorbar(hexbin)
    if log_bins:
        cbar.set_label("log10(Count)")
    else:
        cbar.set_label("Count")
    cbar.ax.tick_params(bottom=False, top=False, left=False, right=False)

    # save and clean up
    pdf.savefig(fig)
    plt.close(fig)


def barcodemap_plot(
    obj, pdf, log=False, bins=DEFAULT_BINS, color=plot_colors["bright3"]
):
    """
    Plot the number of barcodes assigned to each variant.

    Args:
        obj (:py:class:`~enrich2.barcodevariant.BcvSeqLib`): object with counts
            to plot

        pdf (|mpl_PdfPages|): destination file handle

        log (bool): whether to log10-transform the counts

        bins (int): number of histogram bins

        color (str): histogram bar color
    """
    try:
        data = collections.Counter(obj.store["/raw/barcodemap"]["value"])
    except KeyError:
        logger.warning(
            "Failed to generate barcode-variant histogram "
            "(barcode map data not found)"
        )
        return

    if len(list(data.keys())) <= 1:
        logger.warning("Not enough elements to make barcodemap plot")
        return

    # create the plot and set up the axes
    fig, ax = plt.subplots()
    fig.set_tight_layout(True)
    configure_axes(ax)

    # plot the histogram
    ax.hist(list(data.values()), bins=bins, log=log, color=color)

    # set the title and axes labels
    ax.set_title("Barcodes per Variant\n{}".format(obj.name))
    if log:
        ax.set_ylabel("log(Variants)")
    else:
        ax.set_ylabel("Variants")
    ax.set_xlabel("Barcode Count")

    # save and clean up
    pdf.savefig()
    plt.close(fig)


def overlap_merge_plot(obj, pdf):
    """
    Plot the location of all mismatches in the overlap region (split into
    resolved and unresolved) and the location of the first mismatch in the
    overlap region.

    Args:
        obj (OverlapSeqLib): object with counts to plot

        pdf (|mpl_PdfPages|): destination file handle
    """
    # read the data
    try:
        data = obj.store["/raw/overlap_mismatches"]
    except KeyError:
        logger.warning(
            "Failed to generate overlap mismatch plot " "(mismatch data not found)"
        )
        return

    # create the plot and set up the axes
    fig, ax = plt.subplots()
    fig.set_tight_layout(True)
    configure_axes(ax)

    # plotting constants
    width = 0.8  # width of the bars
    xpos = np.arange(len(data.index))  # x values starting at 0

    # plot the stacked barplot for unresolved and resolved mismatches
    ax.bar(
        xpos,
        data["resolved"],
        width,
        color=plot_colors["bright1"],
        label="Resolved Mismatches",
    )
    ax.bar(
        xpos,
        data["unresolved"],
        width,
        bottom=data["resolved"],
        color=plot_colors["bright2"],
        label="Unresolved Mismatches",
    )

    # plot the barplot for first mismatches
    ax.bar(
        xpos,
        data["first"],
        width,
        color=plot_colors["bright3"],
        label="First Mismatches",
    )

    # set title, axes labels, and figure legend
    ax.set_title("Overlap Mismatches\n{}".format(obj.name))
    ax.set_ylabel("Count")
    ax.set_xlabel("Nucleotide Position")
    ax.set_xticks(xpos + width / 2.0)
    ax.set_xticklabels(data.index, rotation=90)
    ax.legend(loc=4)  # bottom-right corner

    # save this figure and set up for next figure
    pdf.savefig(fig)
    plt.close(fig)


def counts_plot(
    seqlib, label, pdf, log=False, bins=DEFAULT_BINS, color=plot_colors["bright1"]
):
    """
    Plot a histogram of processed counts. Excludes counts for wild type and
    synonymous variants.

    Args:
        seqlib (SeqLib): object with counts to plot

        label (str): data label (barcode, variant, etc.)

        pdf (|mpl_PdfPages|): destination file handle

        bins (int): number of histogram bins

        color (str): histogram bar color
    """
    data = seqlib.store.select(
        "/main/{}/counts".format(label),
        where="columns=count & "
        "index != WILD_TYPE_VARIANT & "
        "index != SYNONYMOUS_VARIANT",
    ).values

    if len(data) <= 1:
        logger.warning("Not enough elements to make '{}' counts plot" "".format(label))
        return

    # create the plot and set up the axes
    fig, ax = plt.subplots()
    fig.set_tight_layout(True)
    configure_axes(ax)

    # plot the histogram
    ax.hist(data, bins=bins, log=log, color=color)

    # set title and axes labels
    ax.set_title("{} Counts\n{}".format(label.title(), seqlib.name))
    ax.set_ylabel("{}".format(label.title()))
    if log:
        ax.set_xlabel("log10(Count)")
    else:
        ax.set_xlabel("Count")

    # save figure and clean up
    pdf.savefig()
    plt.close()


def weights_plot(selection, label, pdf):
    """
    Create a boxplot of weights for weighted least squares (WLS) regression 
    scoring. Weights are converted to proportional weights (0 to 1) before 
    plotting.

    *selection* is a :py:class:`~selection.Selection`

    *label* is the data label (barcode, variant, etc.)

    *pdf* is an open PdfPages instance

    """
    # get the data and drop NA values
    data = selection.store.select(
        "/main/{}/weights".format(label),
        where="index != WILD_TYPE_VARIANT & " "index != SYNONYMOUS_VARIANT",
    )
    data.dropna(axis="index", how="any", inplace=True)

    # normalize the weights to [0..1]
    data = data.values / data.values.sum(axis=1)[:, None]

    # create the figure and axes
    fig, ax = plt.subplots()
    fig.set_tight_layout(True)
    configure_axes(ax)

    # add a horizontal line to show the weights if all weights were equal
    ax.axhline(
        y=1.0 / len(selection.timepoints),
        lw=1,
        linestyle="--",
        color=plot_colors["accent1"],
    )

    # plot the boxplot, setting the colors for each part of the plot
    ax.boxplot(
        data,
        notch=False,
        showmeans=True,
        meanline=True,
        sym=None,  # workaround for flierprops being ignored otherwise
        boxprops=dict(color=plot_colors["bright1"]),
        flierprops=dict(markeredgecolor=plot_colors["bright1"], marker="+"),
        medianprops=dict(color=plot_colors["bright2"], linestyle="-"),
        meanprops=dict(color=plot_colors["bright2"], linestyle=":"),
        capprops=dict(color=plot_colors["bright1"], linestyle="-"),
        whiskerprops=dict(color=plot_colors["bright1"], linestyle="--"),
    )

    # add the title and axes labels
    ax.set_title("Regression Weights for {}\n{}".format(label.title(), selection.name))
    ax.set_xlabel("Time Point")
    ax.set_ylabel("Proportional Weight")
    ax.set_xticklabels(selection.timepoints)

    # save the figure and clean up
    pdf.savefig(fig)
    plt.close(fig)


def corr_axes(ax, x, y, score_min, score_max, **kwargs):
    """
    Plot the correlation between scores in two replicates.

    Args:
        ax (matplotlib.axes.Axes): axes object used for plotting

        x (vector): scores for the first replicate

        y (vector): scores for the second replicate

        score_min: minimum score in all plots

        score_max: maximum score in all plots

        **kwargs: arguments to pass to ax.plot for plotting the points (point style, color, etc.)
    """
    configure_axes(ax, xgrid=True)

    # set the axes limits
    ax.set_xlim(left=score_min * 1.05, right=score_max * 1.05)
    ax.set_ylim(bottom=score_min * 1.05, top=score_max * 1.05)

    # plot bolder zero lines
    ax.plot(
        ax.get_xlim(),
        [0.0, 0.0],
        linestyle="-",
        color=plot_colors["accent3"],
        linewidth=1,
    )
    ax.plot(
        [0.0, 0.0],
        ax.get_ylim(),
        linestyle="-",
        color=plot_colors["accent3"],
        linewidth=1,
    )

    # plot the points
    ax.plot(x, y, **kwargs)

    # plot the fit line
    fit = np.polyfit(x, y, 1)
    ax.plot(
        ax.get_xlim(),
        np.poly1d(fit)(ax.get_xlim()),
        linestyle="--",
        color=plot_colors["accent2"],
    )

    # plot the text with regression line info
    r, _ = scipy.stats.pearsonr(x, y)
    ax.text(
        0.05,
        0.95,
        "r^2 = {:.2g}".format(r ** 2),
        horizontalalignment="left",
        verticalalignment="top",
        transform=ax.transAxes,
        fontsize=8,
    )


def hexbin_corr_axes(
    ax, x, y, score_min, score_max, cbar_ax=None, gridsize=50, cmap="YlGnBu_r"
):
    """
    Plot the correlation between scores in two replicates using a hexbin. This 
    function is preferable to corr_axes when plotting a large number of data 
    points.

    Args:
        ax (matplotlib.axes.Axes): axes object used for plotting

        x (vector): scores for the first replicate

        y (vector): scores for the second replicate

        score_min: minimum score in all plots

        score_max: maximum score in all plots

        cbar_ax (matplotlib.axes.Axes): axes object used for plotting the colorbar (optional)

    """
    configure_axes(ax, xgrid=True)

    # set the axes limits
    ax.set_xlim([score_min * 1.05, score_max * 1.05])
    ax.set_ylim([score_min * 1.05, score_max * 1.05])

    # plot bolder zero lines
    ax.plot(
        ax.get_xlim(),
        [0.0, 0.0],
        linestyle="-",
        color=plot_colors["accent3"],
        linewidth=2,
    )
    ax.plot(
        [0.0, 0.0],
        ax.get_ylim(),
        linestyle="-",
        color=plot_colors["accent3"],
        linewidth=2,
    )

    # plot the points
    hb = ax.hexbin(x, y, mincnt=1, cmap=cmap, gridsize=gridsize)
    hb.set_zorder(3)  # on top of the zero lines

    # plot the fit line
    fit = np.polyfit(x, y, 1)
    ax.plot(
        ax.get_xlim(),
        np.poly1d(fit)(ax.get_xlim()),
        linestyle="--",
        linewidth=1,
        color=plot_colors["accent2"],
        zorder=4,
    )

    # plot the text with regression line info
    r, _ = scipy.stats.pearsonr(x, y)
    ax.text(
        0.05,
        0.95,
        "r^2 = {:.3g}".format(r ** 2),
        horizontalalignment="left",
        verticalalignment="top",
        transform=ax.transAxes,
    )

    # plot the colorbar if desired
    if cbar_ax is not None:
        cbar = plt.colorbar(hb, cax=cbar_ax, orientation="horizontal")
        cbar.ax.tick_params(bottom=False, top=False, left=False, right=False)
        cbar.set_label("Count")
        cbar.outline.set_visible(False)

    return hb


def density_ax(ax, ys, xmin, xmax, xlabel, line_params, legend_loc="best"):

    if len(ys) != len(line_params):
        raise ValueError("All y-value sets must have a linestyle")

    configure_axes(ax, xgrid=True)

    d_ys = [KDEUnivariate(y.values) for y in ys]
    [d_y.fit() for d_y in d_ys]

    xs = np.linspace(xmin, xmax, 1000)

    for i in range(len(ys)):
        ax.plot(xs, d_ys[i].evaluate(xs), label=ys[i].name, **line_params[i])

    ax.legend(loc=legend_loc, **legend_params)

    ax.set_xlabel(xlabel, **label_params)
    ax.set_ylabel("Density", **label_params)
    ax.tick_params(axis="both", which="major", **tick_params)



import numpy as np


def rml_estimator(y, sigma2i, iterations=50):
    """Implementation of the robust maximum likelihood estimator.

        ::

            @book{demidenko2013mixed,
              title={Mixed models: theory and applications with R},
              author={Demidenko, Eugene},
              year={2013},
              publisher={John Wiley & Sons}
            }

    """
    w = 1 / sigma2i
    sw = np.sum(w, axis=0)
    beta0 = np.sum(y * w, axis=0) / sw
    sigma2ML = np.sum((y - np.mean(y, axis=0)) ** 2 / (len(beta0) - 1), axis=0)
    eps = np.zeros(beta0.shape)
    betaML = None
    for _ in range(iterations):
        w = 1 / (sigma2i + sigma2ML)
        sw = np.sum(w, axis=0)
        sw2 = np.sum(w ** 2, axis=0)
        betaML = np.sum(y * w, axis=0) / sw
        sigma2ML_new = (
            sigma2ML
            * np.sum(((y - betaML) ** 2) * (w ** 2), axis=0)
            / (sw - (sw2 / sw))
        )
        eps = np.abs(sigma2ML - sigma2ML_new)
        sigma2ML = sigma2ML_new
    var_betaML = 1 / np.sum(1 / (sigma2i + sigma2ML), axis=0)
    return betaML, var_betaML, eps



import logging
import re
from .constants import CODON_TABLE


class WildTypeSequence(object):
    """
    Container class for wild type sequence information. Used by :py:class:`~seqlib.seqlib.VariantSeqLib` objects and
    :py:class:`~enrich2.selection.Selection` or :py:class:`~enrich2.experiment.Experiment` objects that contain
    variant information.

    Requires a *parent_name* that associates this object with a StoreManager object for the
    purposes of error reporting and logging.
    """

    def __init__(self, parent_name):
        self.parent_name = parent_name
        self.dna_seq = None
        self.protein_seq = None
        self.dna_offset = None
        self.protein_offset = None
        self.logger = logging.getLogger("{}.{}".format(__name__, self.__class__))

    def __eq__(self, other):
        # note we don't need to check protein_offset, since it depends on dna_offset and protein_seq
        return (
            self.dna_seq == other.dna_seq
            and self.protein_seq == other.protein_seq
            and self.dna_offset == other.dna_offset
        )

    def __ne__(self, other):
        return not self == other

    def configure(self, cfg):
        try:
            # remove whitespace from WT DNA sequence and capitalize
            self.dna_seq = "".join(cfg["sequence"].split()).upper()

            # check that only valid characters are included (ACGT)
            if not re.match("^[ACGT]+$", self.dna_seq):
                raise ValueError(
                    "WT DNA sequence contains unexpected "
                    "characters [{}]".format(self.parent_name)
                )

            # set the reference offset
            if "reference offset" in cfg:
                try:
                    self.dna_offset = int(cfg["reference offset"])
                except ValueError:
                    raise ValueError(
                        "Invalid reference offset value [{}]".format(self.parent_name)
                    )
            else:
                self.dna_offset = 0

            # handle coding sequences
            if cfg["coding"]:
                # require coding sequences are in-frame
                if len(self.dna_seq) % 3 != 0:
                    raise ValueError(
                        "WT DNA sequence contains incomplete codons [{}]".format(
                            self.parent_name
                        )
                    )

                # perform translation
                self.protein_seq = ""
                for i in range(0, len(self.dna_seq), 3):
                    self.protein_seq += CODON_TABLE[self.dna_seq[i: i + 3]]

                # set the reference offset if it's a multiple of three
                if self.dna_offset % 3 == 0:
                    self.protein_offset = int(self.dna_offset / 3)
                else:
                    self.logger.warning(
                        "Ignoring reference offset for protein changes (not a multiple of three)"
                    )
                    self.protein_offset = 0
            else:
                self.protein_seq = None
                self.protein_offset = None

        except KeyError as key:
            raise KeyError(
                "Missing required config value {key} [{name}]".format(
                    key=key, name=self.parent_name
                )
            )

    def serialize(self):
        """
        Format this object as a config object suitable for dumping to a config file.
        """
        cfg = {
            "sequence": self.dna_seq,
            "coding": self.is_coding(),
            "reference offset": self.dna_offset,
        }
        return cfg

    def is_coding(self):
        return self.protein_seq is not None

    def duplicate(self, new_parent_name):
        """
        Create a copy of this object with the *new_parent_name*.

        Uses the configure and serialize methods to perform the copy.
        """
        new = WildTypeSequence(new_parent_name)
        new.configure(self.serialize())

        if new != self:
            raise ValueError(
                "Failed to duplicate wild type sequence [{}]".format(self.parent_name)
            )
        else:
            return new

    def position_tuples(self, protein=False):
        """
        Return a list of tuples containing the position number (after offset adjustment) and
        single-letter symbol (nucleotide or amino acid) for each position the wild type sequence.
        """
        if protein:
            if not self.is_coding():
                raise AttributeError(
                    "Cannot return wild type protein position tuples for non-coding wild type [{}]".format(
                        self.parent_name
                    )
                )
            else:
                seq = self.protein_seq
                offset = self.protein_offset
        else:
            seq = self.dna_seq
            offset = self.dna_offset

        return [(i + offset + 1, seq[i]) for i in range(len(seq))]
